#!/usr/bin/env drgn
# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Search kernel memory for a byte string. This only works if the kernel was built
with CONFIG_PROC_KCORE=y.
"""

import argparse
import sys

from drgn import Object
from drgn.helpers.common.memory import identify_address
from drgn.helpers.linux.list import list_for_each_entry
from drgn.helpers.linux.mm import for_each_vmap_area, virt_to_page


def virt_to_vmap_address(prog, addr):
    page = virt_to_page(addr)
    for va in for_each_vmap_area(prog):
        vm = va.vm.read_()
        if vm:
            for i, va_page in enumerate(
                Object(
                    prog, prog.array_type(page.type_, vm.nr_pages), address=vm.pages
                ).read_()
            ):
                if va_page == page:
                    return (
                        va.va_start.value_()
                        + (i << prog["PAGE_SHIFT"])
                        + (addr & (prog["PAGE_SIZE"].value_() - 1))
                    )
    return None


def search_memory(prog, needle):
    KCORE_RAM = prog["KCORE_RAM"]
    CHUNK_SIZE = 1024 * 1024
    for kc in list_for_each_entry(
        "struct kcore_list", prog["kclist_head"].address_of_(), "list"
    ):
        if kc.type != KCORE_RAM:
            continue
        start = kc.addr.value_()
        end = start + kc.size.value_()
        for addr in range(start, end, CHUNK_SIZE):
            buf = prog.read(addr, min(CHUNK_SIZE, end - addr))
            i = 0
            while i < len(buf):
                i = buf.find(needle, i)
                if i < 0:
                    break

                vmap_address = virt_to_vmap_address(prog, addr + i)
                if vmap_address is not None:
                    identity = identify_address(prog, vmap_address)
                else:
                    identity = identify_address(prog, addr + i)

                if identity is None:
                    print(hex(addr + i))
                else:
                    print(hex(addr + i), identity)
                i += 1


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Search kernel memory for a byte string"
    )
    parser.add_argument(
        "bytes",
        nargs="?",
        help="hexadecimal bytes to read; if omitted, read byte string from stdin",
    )
    args = parser.parse_args()
    if args.bytes is None:
        needle = sys.stdin.buffer.read()
    else:
        needle = bytes.fromhex(args.bytes)
    search_memory(prog, needle)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Users
-----

The ``drgn.helpers.linux.user`` module provides helpers for working with users
in the Linux kernel.
"""

import operator
from typing import Iterator, Union

from drgn import NULL, IntegerLike, Object, Program
from drgn.helpers.common.prog import takes_program_or_default
from drgn.helpers.linux.list import hlist_for_each_entry

__all__ = (
    "find_user",
    "for_each_user",
)


def _kuid_val(uid: Union[Object, IntegerLike]) -> int:
    if isinstance(uid, Object) and uid.type_.type_name() == "kuid_t":
        uid = uid.val
    return operator.index(uid)


@takes_program_or_default
def find_user(prog: Program, uid: Union[Object, IntegerLike]) -> Object:
    """
    Return the user structure with the given UID.

    :param uid: ``kuid_t`` object or integer.
    :return: ``struct user_struct *`` (``NULL`` if not found)
    """
    try:
        uidhashentry = prog.cache["uidhashentry"]
    except KeyError:
        uidhash_table = prog["uidhash_table"]
        uidhash_sz = len(uidhash_table)
        uidhash_bits = uidhash_sz.bit_length() - 1
        uidhash_mask = uidhash_sz - 1

        def uidhashentry(uid: int) -> Object:
            hash = ((uid >> uidhash_bits) + uid) & uidhash_mask
            return uidhash_table + hash

        prog.cache["uidhashentry"] = uidhashentry

    uid = _kuid_val(uid)
    for user in hlist_for_each_entry(
        "struct user_struct", uidhashentry(uid), "uidhash_node"
    ):
        if user.uid.val == uid:
            return user
    return NULL(prog, "struct user_struct *")


@takes_program_or_default
def for_each_user(prog: Program) -> Iterator[Object]:
    """
    Iterate over all users in the system.

    :return: Iterator of ``struct user_struct *`` objects.
    """
    for hash_entry in prog["uidhash_table"]:
        yield from hlist_for_each_entry(
            "struct user_struct", hash_entry, "uidhash_node"
        )



#!/usr/bin/env drgn
# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Print what is using a struct file *, given as an address."""

import os
import sys

from drgn import Object
from drgn.helpers.linux.fs import for_each_file
from drgn.helpers.linux.list import list_for_each_entry
from drgn.helpers.linux.pid import for_each_task


def find_struct_file_fds(file: Object) -> None:
    for task in for_each_task(file.prog_):
        for fd, fd_file in for_each_file(task):
            if fd_file == file:
                print(
                    f"PID {task.pid.value_()} COMM {task.comm.string_().decode()} FD {fd}"
                )


def find_struct_file_binfmt_misc(file: Object) -> None:
    prog = file.prog_
    for node in list_for_each_entry(
        prog.type("Node", filename="binfmt_misc.c"),
        prog.object("entries", filename="binfmt_misc.c").address_of_(),
        "list",
    ):
        if node.interp_file == file:
            print(f"binfmt_misc {os.fsdecode(node.name.string_())}")


def find_struct_file(file: Object) -> None:
    find_struct_file_fds(file)
    find_struct_file_binfmt_misc(file)


if __name__ == "__main__":
    find_struct_file(Object(prog, "struct file *", int(sys.argv[1], 0)))



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Log Buffer
----------

The ``drgn.helpers.linux.printk`` module provides helpers for reading the Linux
kernel log buffer.
"""

import sys
from typing import TYPE_CHECKING, Dict, List, NamedTuple, Optional, Tuple

if TYPE_CHECKING:
    from _typeshed import SupportsWrite

from drgn import Object, Program, cast, sizeof
from drgn.helpers.common.prog import takes_program_or_default

__all__ = (
    "get_dmesg",
    "get_printk_records",
    "print_dmesg",
)


class PrintkRecord(NamedTuple):
    """Kernel log record."""

    text: bytes
    """Message text."""
    facility: int
    """:manpage:`syslog(3)` facility."""
    level: int
    """Log level."""
    seq: int
    """Sequence number."""
    timestamp: int
    """Timestamp in nanoseconds."""
    caller_tid: Optional[int]
    """
    Thread ID of thread that logged this record, if available.

    This is available if the message was logged from task context and if the
    kernel saves the ``printk()`` caller ID.

    As of Linux 5.10, the kernel always saves the caller ID. From Linux 5.1
    through 5.9, it is saved only if the kernel was compiled with
    ``CONFIG_PRINTK_CALLER``. Before that, it is never saved.
    """
    caller_cpu: Optional[int]
    """
    Processor ID of CPU that logged this record, if available.

    This is available only if the message was logged when not in task context
    (e.g., in an interrupt handler) and if the kernel saves the ``printk()``
    caller ID.

    See :attr:`caller_tid` for when the kernel saves the caller ID.
    """
    continuation: bool
    """Whether this record is a continuation of a previous record."""
    context: Dict[bytes, bytes]
    """
    Additional metadata for the message.

    See the |/dev/kmsg documentation|_ for an explanation of the keys and
    values.

    .. |/dev/kmsg documentation| replace:: ``/dev/kmsg`` documentation
    .. _/dev/kmsg documentation: https://www.kernel.org/doc/Documentation/ABI/testing/dev-kmsg
    """


def _caller_id(caller_id: int) -> Tuple[Optional[int], Optional[int]]:
    if caller_id & 0x80000000:
        return None, caller_id & ~0x80000000
    else:
        return caller_id, None


def _get_printk_records_lockless(prog: Program, prb: Object) -> List[PrintkRecord]:
    ulong_size = sizeof(prog.type("unsigned long"))
    DESC_SV_BITS = ulong_size * 8
    DESC_FLAGS_SHIFT = DESC_SV_BITS - 2
    DESC_FLAGS_MASK = 3 << DESC_FLAGS_SHIFT
    DESC_ID_MASK = DESC_FLAGS_MASK ^ ((1 << DESC_SV_BITS) - 1)

    LOG_CONT = prog["LOG_CONT"].value_()
    desc_committed = prog["desc_committed"].value_()
    desc_finalized = prog["desc_finalized"].value_()

    def record_committed(current_id: int, state_var: int) -> bool:
        state_desc_id = state_var & DESC_ID_MASK
        state = 3 & (state_var >> DESC_FLAGS_SHIFT)
        return (current_id == state_desc_id) and (
            state == desc_committed or state == desc_finalized
        )

    desc_ring = prb.desc_ring
    descs = desc_ring.descs.read_()
    infos = desc_ring.infos.read_()
    desc_ring_mask = (1 << desc_ring.count_bits.value_()) - 1
    text_data_ring = prb.text_data_ring
    text_data_ring_data = text_data_ring.data.read_()
    text_data_ring_mask = (1 << text_data_ring.size_bits) - 1

    result = []

    def add_record(current_id: int) -> None:
        idx = current_id & desc_ring_mask
        desc = descs[idx].read_()
        if not record_committed(current_id, desc.state_var.counter.value_()):
            return

        lpos_begin = desc.text_blk_lpos.begin & text_data_ring_mask
        lpos_next = desc.text_blk_lpos.next & text_data_ring_mask
        lpos_begin += ulong_size

        if lpos_begin == lpos_next:
            # Data-less record.
            return
        if lpos_begin > lpos_next:
            # Data wrapped.
            lpos_begin -= lpos_begin
        info = infos[idx].read_()
        text_len = info.text_len
        if lpos_next - lpos_begin < text_len:
            # Truncated record.
            text_len = lpos_next - lpos_begin

        caller_tid, caller_cpu = _caller_id(info.caller_id.value_())

        context = {}
        subsystem = info.dev_info.subsystem.string_()
        device = info.dev_info.device.string_()
        if subsystem:
            context[b"SUBSYSTEM"] = subsystem
        if device:
            context[b"DEVICE"] = device

        result.append(
            PrintkRecord(
                text=prog.read(text_data_ring_data + lpos_begin, text_len),
                facility=info.facility.value_(),
                level=info.level.value_(),
                seq=info.seq.value_(),
                timestamp=info.ts_nsec.value_(),
                caller_tid=caller_tid,
                caller_cpu=caller_cpu,
                continuation=bool(info.flags.value_() & LOG_CONT),
                context=context,
            )
        )

    head_id = desc_ring.head_id.counter.value_()
    current_id = desc_ring.tail_id.counter.value_()
    while current_id != head_id:
        add_record(current_id)
        current_id = (current_id + 1) & DESC_ID_MASK
    add_record(current_id)
    return result


def _get_printk_records_structured(prog: Program) -> List[PrintkRecord]:
    try:
        printk_logp_type = prog.type("struct printk_log *")
    except LookupError:
        # Before Linux kernel commit 62e32ac3505a ("printk: rename struct log
        # to struct printk_log") (in v3.11), records were "struct log" instead
        # of "struct printk_log". RHEL 7 kernel still uses old naming.
        printk_logp_type = prog.type("struct log *")

    have_caller_id = printk_logp_type.type.has_member("caller_id")
    LOG_CONT = prog["LOG_CONT"].value_()

    result = []
    # Between Linux kernel commits cbd357008604 ("bpf: verifier (add ability to
    # receive verification log)") (in v3.18) and e7bf8249e8f1 ("bpf:
    # encapsulate verifier log state into a structure") (in v4.15),
    # kernel/bpf/verifier.c also contains a variable named log_buf.
    log_buf = prog.object("log_buf", filename="printk.c").read_()
    current_idx = prog["log_first_idx"].read_()
    next_idx = prog["log_next_idx"].read_()
    seq = prog["log_first_seq"].value_()
    while current_idx != next_idx:
        logp = cast(printk_logp_type, log_buf + current_idx)
        log = logp[0].read_()
        text_len = log.text_len.value_()
        dict_len = log.dict_len.value_()
        text_dict = prog.read(logp + 1, text_len + dict_len)

        if have_caller_id:
            caller_tid, caller_cpu = _caller_id(log.caller_id.value_())
        else:
            caller_tid = caller_cpu = None

        context = {}
        if dict_len:
            for elmt in text_dict[text_len:].split(b"\0"):
                key, value = elmt.split(b"=", 1)
                context[key] = value

        result.append(
            PrintkRecord(
                text=text_dict[:text_len],
                facility=log.facility.value_(),
                level=log.level.value_(),
                seq=seq,
                timestamp=log.ts_nsec.value_(),
                caller_tid=caller_tid,
                caller_cpu=caller_cpu,
                continuation=bool(log.flags.value_() & LOG_CONT),
                context=context,
            )
        )
        log_len = log.len.read_()
        if log_len:
            current_idx += log_len
        else:
            # Zero means the buffer wrapped around.
            if current_idx < next_idx:
                # Avoid getting into an infinite loop if the buffer is
                # corrupted.
                break
            current_idx -= current_idx
        seq += 1
    return result


@takes_program_or_default
def get_printk_records(prog: Program) -> List[PrintkRecord]:
    """Get a list of records in the kernel log buffer."""
    # Linux kernel commit 896fbe20b4e2 ("printk: use the lockless ringbuffer")
    # (in v5.10) changed the ring buffer structure completely.
    try:
        prb = prog["prb"]
    except KeyError:
        return _get_printk_records_structured(prog)
    else:
        return _get_printk_records_lockless(prog, prb)


@takes_program_or_default
def get_dmesg(prog: Program) -> bytes:
    """
    Get the contents of the kernel log buffer formatted like
    :manpage:`dmesg(1)`.

    If you just want to print the log buffer, use :func:`print_dmesg()`.

    The format of each line is:

    .. code-block::

        [   timestamp] message

    If you need to format the log buffer differently, use
    :func:`get_printk_records()` and format it yourself.
    """
    lines = [
        b"[% 5d.%06d] %s"
        % (
            record.timestamp // 1000000000,
            record.timestamp % 1000000000 // 1000,
            record.text,
        )
        for record in get_printk_records(prog)
    ]
    lines.append(b"")  # So we get a trailing newline.
    return b"\n".join(lines)


@takes_program_or_default
def print_dmesg(prog: Program, *, file: "Optional[SupportsWrite[str]]" = None) -> None:
    """
    Print the contents of the kernel log buffer.

    >>> print_dmesg()
    [    0.000000] Linux version 6.8.0-vmtest28.1default (drgn@drgn) (x86_64-linux-gcc (GCC) 12.2.0, GNU ld (GNU Binutils) 2.39) #1 SMP PREEMPT_DYNAMIC Mon Mar 11 06:38:45 UTC 2024
    [    0.000000] Command line: rootfstype=9p rootflags=trans=virtio,cache=loose,msize=1048576 ro console=ttyS0,115200 panic=-1 crashkernel=256M init=/tmp/drgn-vmtest-rudzppeo/init
    [    0.000000] BIOS-provided physical RAM map:
    ...

    :param file: File to print to. Defaults to :data:`sys.stdout`.
    """
    (sys.stdout if file is None else file).write(
        get_dmesg(prog).decode(errors="replace")
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Print registered platform drivers."""

from drgn import NULL, container_of
from drgn.helpers.linux.list import list_for_each_entry


def bus_to_subsys(bus):
    for sp in list_for_each_entry(
        "struct subsys_private",
        prog["bus_kset"].list.address_of_(),
        "subsys.kobj.entry",
    ):
        if sp.bus == bus:
            return sp
    return NULL(bus.prog_, "struct subsys_private *")


sp = bus_to_subsys(prog["platform_bus_type"].address_of_())
for priv in list_for_each_entry(
    "struct driver_private", sp.drivers_kset.list.address_of_(), "kobj.entry"
):
    driver = priv.driver
    print(driver.name.string_().decode())
    platform_driver = container_of(driver, "struct platform_driver", "driver")
    print(platform_driver)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Kernel Threads
--------------

The ``drgn.helpers.linux.kthread`` module provides helpers for working with
Linux kernel threads, a.k.a. kthreads.
"""

from drgn import Object, cast, container_of

__all__ = (
    "kthread_data",
    "to_kthread",
)


def to_kthread(task: Object) -> Object:
    """
    Get the kthread information for a task.

    >>> to_kthread(find_task(3))
    *(struct kthread *)0xffff8ef600191580 = {
            ...
            .threadfn = (int (*)(void *))kthread_worker_fn+0x0 = 0xffffffffba1e61b0,
            .full_name = (char *)0xffff8ef6003d4ac0 = "pool_workqueue_release",
    }

    :param task: ``struct task *``
    :return: ``struct kthread *``
    """
    try:
        # Since Linux kernel commit e32cf5dfbe22 ("kthread: Generalize
        # pf_io_worker so it can point to struct kthread") (in v5.17), the
        # struct kthread * is in task->worker_private.
        return cast("struct kthread *", task.worker_private)
    except AttributeError:
        if "free_kthread_struct" in task.prog_:
            # Between that and Linux kernel commit 1da5c46fa965 ("kthread: Make
            # struct kthread kmalloc'ed") (in v4.10), it is in
            # task->set_child_tid. Unfortunately we can only distinguish this
            # by looking for another function added in that commit.
            return cast("struct kthread *", task.set_child_tid)
        else:
            # Before that, task->vfork_done points to kthread->exited.
            return container_of(task.vfork_done, "struct kthread", "exited")


def kthread_data(task: Object) -> Object:
    """
    Get the data that was specified when a kthread was created.

    >>> kthread_data(find_task(3))
    (void *)0xffff8ef6001812c0

    :param task: ``struct task *``
    :return: ``void *``
    """
    return to_kthread(task).data.read_()



# Copyright (c) ByteDance, Inc. and its affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Traffic Control (TC)
--------------------

The ``drgn.helpers.linux.tc`` module provides helpers for working with the
Linux kernel Traffic Control (TC) subsystem.
"""

import operator

from drgn import NULL, IntegerLike, Object
from drgn.helpers.linux.list import hlist_for_each_entry, list_for_each_entry

__all__ = ("qdisc_lookup",)


def qdisc_lookup(dev: Object, major: IntegerLike) -> Object:
    """
    Get a Qdisc from a device and a major handle number.  It is worth noting
    that conventionally handles are hexadecimal, e.g. ``10:`` in a ``tc``
    command means major handle 0x10.

    :param dev: ``struct net_device *``
    :param major: Qdisc major handle number.
    :return: ``struct Qdisc *`` (``NULL`` if not found)
    """
    major = operator.index(major) << 16

    roots = [dev.qdisc]
    if dev.ingress_queue:
        roots.append(dev.ingress_queue.qdisc_sleeping)

    # Since Linux kernel commit 59cc1f61f09c ("net: sched: convert qdisc linked
    # list to hashtable") (in v4.7), a device's child Qdiscs are maintained in
    # a hashtable in its struct net_device. Before that, they are maintained in
    # a linked list in their root Qdisc.
    use_hashtable = dev.prog_.type("struct net_device").has_member("qdisc_hash")

    for root in roots:
        if root.handle == major:
            return root

        if use_hashtable:
            for head in root.dev_queue.dev.qdisc_hash:
                for qdisc in hlist_for_each_entry(
                    "struct Qdisc", head.address_of_(), "hash"
                ):
                    if qdisc.handle == major:
                        return qdisc
        else:
            for qdisc in list_for_each_entry(
                "struct Qdisc", root.list.address_of_(), "list"
            ):
                if qdisc.handle == major:
                    return qdisc

    return NULL(dev.prog_, "struct Qdisc *")



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Kernfs
------

The ``drgn.helpers.linux.kernfs`` module provides helpers for working with the
kernfs pseudo filesystem interface in :linux:`include/linux/kernfs.h`.
"""

import os

from drgn import NULL, Object, Path
from drgn.helpers.linux.rbtree import rbtree_inorder_for_each_entry

__all__ = (
    "kernfs_name",
    "kernfs_parent",
    "kernfs_path",
    "kernfs_root",
    "kernfs_walk",
)


def kernfs_root(kn: Object) -> Object:
    """
    Get the kernfs root that the given kernfs node belongs to.

    :param kn: ``struct kernfs_node *``
    :return: ``struct kernfs_root *``
    """
    knp = kernfs_parent(kn)
    if knp:
        kn = knp
    return kn.dir.root.read_()


def kernfs_parent(kn: Object) -> Object:
    """
    Get the parent of the given kernfs node.

    :param kn: ``struct kernfs_node *``
    :return: ``struct kernfs_node *``
    """
    # Linux kernel commit 633488947ef6 ("kernfs: Use RCU to access
    # kernfs_node::parent.") (in v6.15) renamed the parent member.
    try:
        return kn.__parent.read_()
    except AttributeError:
        return kn.parent.read_()


def kernfs_name(kn: Object) -> bytes:
    """
    Get the name of the given kernfs node.

    :param kn: ``struct kernfs_node *``
    """
    if not kn:
        return b"(null)"
    return kn.name.string_() if kernfs_parent(kn) else b"/"


def kernfs_path(kn: Object) -> bytes:
    """
    Get full path of the given kernfs node.

    :param kn: ``struct kernfs_node *``
    """
    if not kn:
        return b"(null)"

    root_kn = kernfs_root(kn).kn
    if kn == root_kn:
        return b"/"

    names = []
    while kn != root_kn:
        names.append(kn.name.string_())
        kn = kernfs_parent(kn)
    names.append(root_kn.name.string_())
    names.reverse()

    return b"/".join(names)


def kernfs_walk(parent: Object, path: Path) -> Object:
    """
    Find the kernfs node with the given path from the given parent kernfs node.

    :param parent: ``struct kernfs_node *``
    :param path: Path name.
    :return: ``struct kernfs_node *`` (``NULL`` if not found)
    """
    kernfs_nodep_type = parent.type_
    kernfs_node_type = kernfs_nodep_type.type
    for name in os.fsencode(path).split(b"/"):
        if not name:
            continue

        for parent in rbtree_inorder_for_each_entry(
            kernfs_node_type, parent.dir.children.address_of_(), "rb"
        ):
            if (
                parent.name.string_() == name
                and not parent.ns  # For now, we don't bother with namespaced kernfs nodes.
            ):
                break
        else:
            return NULL(parent.prog_, kernfs_nodep_type)
    return parent



#!/usr/bin/env drgn
# Copyright (c) SUSE Linux.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Display system information and configuration data."""

from datetime import datetime
from datetime import timedelta

from drgn.helpers.common.format import number_in_binary_units
from drgn.helpers.linux import for_each_online_cpu
from drgn.helpers.linux.mm import totalram_pages
from drgn.helpers.linux.pid import for_each_task
from drgn.helpers.linux.sched import loadavg


def print_line(key, value):
    print(f"{key:<16} {value}")


uts = prog["init_uts_ns"].name

timekeeper = prog["shadow_timekeeper"]
date = datetime.fromtimestamp(timekeeper.xtime_sec).strftime("%c")
uptime = timedelta(seconds=timekeeper.ktime_sec.value_())
load = ", ".join([f"{v:.2f}" for v in loadavg(prog)])
totalram = (prog['PAGE_SIZE'] * totalram_pages(prog)).value_()


print_line("CPUS", len(list(for_each_online_cpu(prog))))
print_line("DATE", date)
print_line("UPTIME", uptime)
print_line("LOAD AVERAGE", load)
print_line("TASKS", len(list(for_each_task(prog))))
print_line("NODENAME", uts.nodename.string_().decode())
print_line("RELEASE", uts.release.string_().decode())
print_line("VERSION", uts.version.string_().decode())
print_line("MACHINE", uts.machine.string_().decode())
print_line("MEMORY", number_in_binary_units(totalram))



#!/usr/bin/env drgn
# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""List BPF programs or maps and their properties unavailable via kernel API."""

import sys
import drgn
import argparse

from drgn import container_of
from drgn.helpers.common.type import enum_type_to_class
from drgn.helpers.linux import (
    bpf_map_for_each,
    bpf_prog_for_each,
    bpf_link_for_each,
    list_for_each_entry,
    hlist_for_each_entry,
)

BpfMapType = enum_type_to_class(prog.type("enum bpf_map_type"), "BpfMapType")
BpfProgType = enum_type_to_class(prog.type("enum bpf_prog_type"), "BpfProgType")
BpfAttachType = enum_type_to_class(prog.type("enum bpf_attach_type"), "BpfAttachType")
BpfLinkType = enum_type_to_class(prog.type("enum bpf_link_type"), "BpfLinkType")


def bpf_attach_type_to_tramp(attach_type):
    # bpf_tramp_prog_type is available since linux kernel 5.5, this code should
    # be called only after checking for bpf_prog.aux.trampoline to be present
    # though so no error checking here.
    BpfProgTrampType = enum_type_to_class(
        prog.type("enum bpf_tramp_prog_type"), "BpfProgTrampType"
    )

    at = BpfAttachType(attach_type)

    if at == BpfAttachType.BPF_TRACE_FENTRY:
        return BpfProgTrampType.BPF_TRAMP_FENTRY

    if at == BpfAttachType.BPF_TRACE_FEXIT:
        return BpfProgTrampType.BPF_TRAMP_FEXIT

    return BpfProgTrampType.BPF_TRAMP_REPLACE


class BpfTramp(object):
    def __init__(self, tr):
        self.tr = tr

    def get_progs(self):
        if not self.tr:
            return

        if self.tr.extension_prog:
            yield self.tr.extension_prog
            return

        try:
            for head in self.tr.progs_hlist:
                for tramp_aux in hlist_for_each_entry(
                    "struct bpf_prog_aux", head, "tramp_hlist"
                ):
                    yield tramp_aux.prog
        except LookupError:
            return


class BpfMap(object):
    def __init__(self, bpf_map):
        self.map = bpf_map

    @staticmethod
    def inspect_owner(owner):
        type_ = BpfProgType(owner.type).name
        jited = " JITed" if owner.jited.value_() else ""
        return f"{type_:32}{jited}"

    def get_owner(self):
        try:
            owner = self.map.member_("owner")
            return self.inspect_owner(owner)
        except LookupError:
            return ""

    def __repr__(self):
        id_ = self.map.id.value_()
        type_ = BpfMapType(self.map.map_type).name
        name = self.map.name.string_().decode()

        return f"{id_:>6}: {type_:32} {name:32}"


class BpfProg(object):
    def __init__(self, bpf_prog):
        self.prog = bpf_prog

    def is_subprog(self):
        return self.prog.aux.func_idx.value_() != 0

    @staticmethod
    def __get_btf_name(btf, btf_id):
        type_ = btf.types[btf_id]
        if type_.name_off < btf.hdr.str_len:
            return btf.strings[type_.name_off].address_of_().string_().decode()
        return ""

    def get_btf_name(self):
        aux = self.prog.aux
        if aux.btf:
            # func_info[0] points to BPF program function itself.
            return self.__get_btf_name(aux.btf, aux.func_info[0].type_id)
        return ""

    def get_ksym_name(self):
        try:
            ksym = self.prog.aux.member_("ksym")
            return ksym.name.string_().decode()[26:]
        except LookupError:
            return ""

    def get_prog_name(self):
        if self.is_subprog():
            return self.get_ksym_name() or self.prog.aux.name.string_().decode()
        return self.get_btf_name() or self.prog.aux.name.string_().decode()

    def get_used_maps(self):
        for i in range(0, self.prog.aux.used_map_cnt.value_()):
            yield BpfMap(self.prog.aux.used_maps[i])

    def get_subprogs(self):
        for i in range(0, self.prog.aux.func_cnt.value_()):
            yield i, BpfProg(self.prog.aux.func[i])

    def get_linked_func(self):
        kind = bpf_attach_type_to_tramp(self.prog.expected_attach_type)

        linked_prog = self.prog.aux.linked_prog
        linked_prog_id = linked_prog.aux.id.value_()
        linked_btf_id = self.prog.aux.attach_btf_id.value_()

        linked_name = (
            f"{BpfProg(linked_prog).get_prog_name()}->"
            f"{self.__get_btf_name(linked_prog.aux.btf, linked_btf_id)}()"
        )

        return f"{linked_prog_id}->{linked_btf_id}: {kind.name} {linked_name}"

    def get_attach_func(self):
        try:
            func_ = self.prog.aux.attach_func_name
            if func_:
                return func_.string_().decode()
        except LookupError:
            pass

        return ""

    def get_tramp_progs(self):
        try:
            # Trampoline was changed to dst_trampoline since Linux kernel commit
            # 3aac1ead5eb6 ("bpf: Move prog->aux->linked_prog and trampoline into
            # bpf_link on attach") (in v5.10).
            # Try to get dst_trampoline first.
            tr = self.prog.aux.member_("dst_trampoline")
        except LookupError:
            tr = None

        try:
            tr = self.prog.aux.member_("trampoline") if not tr else tr
        except LookupError:
            # Trampoline is available since Linux kernel commit
            # fec56f5890d9 ("bpf: Introduce BPF trampoline") (in v5.5).
            # Skip trampoline if current kernel doesn't support it.
            return

        return BpfTramp(tr).get_progs()

    def __repr__(self):
        id_ = self.prog.aux.id.value_()
        type_ = BpfProgType(self.prog.type).name
        name = self.get_prog_name()
        try:
            tail_call_reachable = self.prog.aux.member_("tail_call_reachable").value_()
        except LookupError:
            tail_call_reachable = None

        tail_call_desc = " tail_call_reachable" if tail_call_reachable else ""

        return f"{id_:>6}: {type_:32} {name:32}{tail_call_desc}"


def list_bpf_progs(show_details=False):
    for bpf_prog_ in bpf_prog_for_each(prog):
        bpf_prog = BpfProg(bpf_prog_)
        print(f"{bpf_prog}")

        if not show_details:
            continue

        linked_progs = bpf_prog.get_tramp_progs()
        if linked_progs:
            for linked_prog in linked_progs:
                print(f"\tlinked: {BpfProg(linked_prog)}")

        for map_ in bpf_prog.get_used_maps():
            print(f"\t{'used map:':9} {map_}")

        for index, subprog in bpf_prog.get_subprogs():
            print(f"\t{f'func[{index:>2}]:':9} {subprog}")


def __list_bpf_progs(args):
    list_bpf_progs(args.show_details)


class BpfProgArrayMap(BpfMap):
    def __init__(self, bpf_map):
        super().__init__(bpf_map)
        self.prog_array = container_of(bpf_map, "struct bpf_array", "map")

    def get_owner(self):
        try:
            owner = self.prog_array.aux.member_("owner")
            return super().inspect_owner(owner)
        except LookupError:
            return ""

    def get_prog_array(self):
        for i in range(0, self.map.max_entries):
            prog_ = self.prog_array.ptrs[i]
            if prog_:
                yield i, drgn.cast("struct bpf_prog *", prog_)

    def get_poke_progs(self):
        for poke in list_for_each_entry(
            "struct prog_poke_elem",
            self.prog_array.aux.poke_progs.address_of_(),
            "list",
        ):
            yield poke.aux.prog

    def __repr__(self):
        owner = self.get_owner()
        owner = super().get_owner() if not owner else owner
        array = self.get_prog_array()
        poke_progs = self.get_poke_progs()

        owner_str = f"{'owner:':9} {owner}" if owner else ""
        array_str = (
            "\n\t".join(
                f"{f'idx[{index:>3}]:':9} {BpfProg(prog)}" for index, prog in array
            )
            if array
            else ""
        )
        poke_progs_str = (
            "\n\t".join(f"{'poke:':9} {BpfProg(poke)}" for poke in poke_progs)
            if poke_progs
            else ""
        )

        return "\n\t".join(x for x in [owner_str, array_str, poke_progs_str] if x)


def show_bpf_map_details(bpf_map):
    if bpf_map.map_type == BpfMapType.BPF_MAP_TYPE_PROG_ARRAY:
        r = BpfProgArrayMap(bpf_map).__repr__()
    else:
        r = None

    if r:
        print(f"\t{r}")


def list_bpf_maps(show_details=False):
    for map_ in bpf_map_for_each(prog):
        bpf_map = BpfMap(map_)
        print(f"{bpf_map}")

        if show_details:
            show_bpf_map_details(map_)


def __list_bpf_maps(args):
    list_bpf_maps(args.show_details)


class BpfLink(object):
    def __init__(self, bpf_link):
        self.link = bpf_link

    def __repr__(self):
        id_ = self.link.id.value_()
        type_ = BpfLinkType(self.link.type).name

        return f"{id_:>6}: {type_:32}"


class BpfTracingLink(BpfLink):
    def __init__(self, link):
        super().__init__(link)
        self.tracing = drgn.cast("struct bpf_tracing_link *", link)

    def get_tgt_prog(self):
        return self.tracing.tgt_prog

    def get_linked_progs(self):
        return BpfTramp(self.tracing.trampoline).get_progs()

    def __repr__(self):
        tgt_prog = self.get_tgt_prog()
        linked_progs = self.get_linked_progs()

        tgt_prog_str = f"target: {BpfProg(tgt_prog)}" if tgt_prog else ""
        linked_progs_str = (
            "\n".join(f"linked: {BpfProg(linked_prog)}" for linked_prog in linked_progs)
            if linked_progs
            else ""
        )

        return "\n\t".join(x for x in [tgt_prog_str, linked_progs_str] if x)


class BpfXdpLink(BpfLink):
    def __init__(self, link):
        super().__init__(link)
        self.xdp = drgn.cast("struct bpf_xdp_link *", link)

    def get_dev(self):
        return self.xdp.dev

    XDP_FLAGS_SKB_MODE = 1 << 1
    XDP_FLAGS_DRV_MODE = 1 << 2
    XDP_FLAGS_HW_MODE = 1 << 3

    def get_mode(self):
        flags = self.xdp.flags.value_()
        if flags & self.XDP_FLAGS_HW_MODE:
            return "HARDWARE"
        if flags & self.XDP_FLAGS_DRV_MODE:
            return "DRIVER"
        if flags & self.XDP_FLAGS_SKB_MODE:
            return "GENERIC"
        return "UNKNOWN"

    def __repr__(self):
        dev = self.get_dev()
        mode = self.get_mode()

        ifname, ifindex = dev.name.string_().decode(), dev.ifindex.value_()
        return f"{'netdev:':<9} {ifname}({ifindex})" + f"\n\t{'mode:':<9} {mode}"


def show_bpf_link_details(link):
    if link.type == BpfLinkType.BPF_LINK_TYPE_TRACING:
        r = BpfTracingLink(link).__repr__()
    elif link.type == BpfLinkType.BPF_LINK_TYPE_XDP:
        r = BpfXdpLink(link).__repr__()
    else:
        r = None

    if r:
        print(f"\t{r}")


def list_bpf_links(show_details=False):
    for link in bpf_link_for_each(prog):
        bpf_link = BpfLink(link)
        print(f"{bpf_link}")

        bpf_prog = BpfProg(link.prog)
        print(f"\tprog:   {bpf_prog}")

        attach_func = bpf_prog.get_attach_func()
        if attach_func:
            print(f"\tattach:   {attach_func}")

        if show_details:
            show_bpf_link_details(link)


def __list_bpf_links(args):
    list_bpf_links(args.show_details)


def __run_interactive(args):
    try:
        from drgn.cli import run_interactive
    except ImportError:
        sys.exit("Interactive mode requires drgn 0.0.23+")

    def should_add_to_globals(name):
        if name.startswith("__"):
            return False
        return "bpf" in name or "Bpf" in name or "btf" in name

    globals_keys = globals().keys()

    def globals_func(globals_):
        for key in globals_keys:
            if should_add_to_globals(key):
                globals_[key] = globals()[key]

        return globals_

    run_interactive(prog, globals_func=globals_func)


def main():
    parser = argparse.ArgumentParser(
        description="drgn script to list BPF programs or maps and their properties unavailable via kernel API"
    )

    subparsers = parser.add_subparsers(title="subcommands", dest="subcommand")
    subparsers.required = True

    prog_parser = subparsers.add_parser("prog", aliases=["p"], help="list BPF programs")
    prog_parser.set_defaults(func=__list_bpf_progs)
    prog_parser.add_argument(
        "--show-details", action="store_true", help="show program internal details"
    )

    map_parser = subparsers.add_parser("map", aliases=["m"], help="list BPF maps")
    map_parser.set_defaults(func=__list_bpf_maps)
    map_parser.add_argument(
        "--show-details", action="store_true", help="show map internal details"
    )

    link_parser = subparsers.add_parser("link", aliases=["l"], help="list BPF links")
    link_parser.set_defaults(func=__list_bpf_links)
    link_parser.add_argument(
        "--show-details", action="store_true", help="show link internal details"
    )

    interact_parser = subparsers.add_parser(
        "interact", aliases=["i"], help="start interactive shell, requires 0.0.23+ drgn"
    )
    interact_parser.set_defaults(func=__run_interactive)

    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



# Copyright (c) 2022, Oracle and/or its affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Lockless Lists
--------------

The ``drgn.helpers.linux.llist`` module provides helpers for working with the
lockless, ``NULL``-terminated, singly-linked list implementation in
:linux:`include/linux/llist.h` (``struct llist_head`` and ``struct
llist_node``).
"""

from typing import Iterator, Union

from drgn import NULL, Object, Type, container_of

__all__ = (
    "llist_empty",
    "llist_first_entry",
    "llist_first_entry_or_null",
    "llist_for_each",
    "llist_for_each_entry",
    "llist_is_singular",
    "llist_next_entry",
)


def llist_empty(head: Object) -> bool:
    """
    Return whether an llist is empty.

    :param head: ``struct llist_head *``
    """
    return not head.first


def llist_is_singular(head: Object) -> bool:
    """
    Return whether an llist has only one element.

    :param head: ``struct llist_head *``
    """
    first = head.first.read_()
    return bool(first) and not first.next


def llist_first_entry(head: Object, type: Union[str, Type], member: str) -> Object:
    """
    Return the first entry in an llist.

    The list is assumed to be non-empty.

    See also :func:`llist_first_entry_or_null()`.

    :param head: ``struct llist_head *``
    :param type: Entry type.
    :param member: Name of ``struct llist_node`` member in entry type.
    :return: ``type *``
    """
    return container_of(head.first, type, member)


def llist_first_entry_or_null(
    head: Object, type: Union[str, Type], member: str
) -> Object:
    """
    Return the first entry in an llist or ``NULL`` if the llist is empty.

    See also :func:`llist_first_entry()`.

    :param head: ``struct llist_head *``
    :param type: Entry type.
    :param member: Name of ``struct llist_node`` member in entry type.
    :return: ``type *``
    """
    first = head.first.read_()
    if first:
        return container_of(first, type, member)
    else:
        return NULL(head.prog_, head.prog_.pointer_type(head.prog_.type(type)))


def llist_next_entry(pos: Object, member: str) -> Object:
    """
    Return the next entry in an llist.

    :param pos: ``type*``
    :param member: Name of ``struct llist_node`` member in entry type.
    :return: ``type *``
    """
    return container_of(getattr(pos, member).next, pos.type_.type, member)


def llist_for_each(node: Object) -> Iterator[Object]:
    """
    Iterate over all of the nodes in an llist starting from a given node.

    :param node: ``struct llist_node *``
    :return: Iterator of ``struct llist_node *`` objects.
    """
    pos = node.read_()
    while pos:
        yield pos
        pos = pos.next.read_()


def llist_for_each_entry(
    type: Union[str, Type], node: Object, member: str
) -> Iterator[Object]:
    """
    Iterate over all of the entries in an llist starting from a given node.

    :param type: Entry type.
    :param node: ``struct llist_node *``
    :param member: Name of ``struct llist_node`` member in entry type.
    :return: Iterator of ``type *`` objects.
    """
    type = node.prog_.type(type)
    for pos in llist_for_each(node):
        yield container_of(pos, type, member)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Helpers
-------

The ``drgn.helpers`` package contains subpackages which provide helpers for
working with particular types of programs. Currently, there are common helpers
and helpers for the Linux kernel. In the future, there may be helpers for,
e.g., glibc and libstdc++.
"""


class ValidationError(Exception):
    """
    Error raised by a :ref:`validator <validators>` when an inconsistent or
    invalid state is detected.
    """



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Stack
-----

The ``drgn.helpers.common.stack`` module provides helpers for working with
stack traces.
"""

from typing import Any, Dict

from drgn import FaultError, PlatformFlags, StackTrace
from drgn.helpers.common.memory import identify_address

__all__ = ("print_annotated_stack",)


def print_annotated_stack(trace: StackTrace) -> None:
    """
    Print the contents of stack memory in a stack trace, annotating values that
    can be identified.

    Currently, this will identify any addresses on the stack with
    :func:`~drgn.helpers.common.memory.identify_address()`.

    See :func:`~drgn.helpers.common.memory.print_annotated_memory()` for a
    similar function that annotates arbitrary memory ranges.

    >>> print_annotated_stack(stack_trace(1))
    STACK POINTER     VALUE
    [stack frame #0 at 0xffffffff8dc93c41 (__schedule+0x429/0x488) in context_switch at ./kernel/sched/core.c:5209:2 (inlined)]
    [stack frame #1 at 0xffffffff8dc93c41 (__schedule+0x429/0x488) in __schedule at ./kernel/sched/core.c:6521:8]
    ffffa903c0013d28: ffffffff8d8497bf [function symbol: __flush_tlb_one_user+0x5]
    ffffa903c0013d30: 000000008d849eb5
    ffffa903c0013d38: 0000000000000001
    ffffa903c0013d40: 0000000000000004
    ffffa903c0013d48: efdea37bb7cb1f00
    ffffa903c0013d50: ffff926641178000 [slab object: task_struct+0x0]
    ffffa903c0013d58: ffff926641178000 [slab object: task_struct+0x0]
    ffffa903c0013d60: ffffa903c0013e10
    ffffa903c0013d68: ffff926641177ff0 [slab object: mm_struct+0x70]
    ffffa903c0013d70: ffff926641178000 [slab object: task_struct+0x0]
    ffffa903c0013d78: ffff926641178000 [slab object: task_struct+0x0]
    ffffa903c0013d80: ffffffff8dc93d29 [function symbol: schedule+0x89]
    ...

    :param trace: Stack trace to print.
    """
    prog = trace.prog

    # platform must be known if there is a stack trace
    assert prog.platform is not None

    if prog.platform.flags & PlatformFlags.IS_LITTLE_ENDIAN:
        byteorder = "little"
    else:
        byteorder = "big"

    if prog.platform.flags & PlatformFlags.IS_64_BIT:
        word_size = 8
        line_format = "{:016x}: {:016x}{}"
        print("STACK POINTER     VALUE")
    else:
        word_size = 4
        line_format = "{:08x}: {:08x}{}"
        print("STACK     VALUE\nPOINTER")

    cache: Dict[Any, Any] = {}
    start = 0
    while start < len(trace):
        # Find the bounds of this stack. Our heuristics for the end of the
        # stack are:
        #
        # 1. An interrupted frame.
        # 2. A frame with a stack pointer less than the previous frame's stack
        #    pointer (since the stack grows down in all of the architectures we
        #    support).
        # 3. A frame with a stack pointer much greater than the previous
        #    frame's stack pointer. Our arbitrary threshold is 128 MB. (Linux
        #    kernel stacks are at most 16 KB as of Linux 6.8, and userspace
        #    stacks are limited to 8 MB by default, so this threshold could be
        #    adjusted if needed.)
        end = start + 1
        while (
            end < len(trace)
            and not trace[end].interrupted
            and 0 <= trace[end].sp - trace[end - 1].sp <= 128 * 1024 * 1024
        ):
            end += 1

        # Gather the frames for this stack.
        frames = [trace[i] for i in range(start, end)]
        frames_addrs = [frame.sp for frame in frames]

        start_addr = frames_addrs[0]
        end_addr = frames_addrs[-1] + word_size - 1
        stack_size = end_addr - start_addr + 1

        try:
            stack_bytes = prog.read(start_addr, stack_size)
        except FaultError:
            # Couldn't read the stack. Just print the frames.
            for frame in frames:
                print(f"[stack frame {frame}]")

            start = end
            continue

        frame_ind = 0

        for offset in range(0, len(stack_bytes), word_size):
            addr = start_addr + offset

            word_bytes = stack_bytes[offset : offset + word_size]
            word_val = int.from_bytes(
                word_bytes,
                # The byteorder parameter is annotated as
                # Literal['little', 'big'], but mypy infers that our byteorder
                # variable is str.
                byteorder=byteorder,  # type: ignore[arg-type]
            )

            # There may be multiple frames matching this address (usually
            # because of inline frames).
            while frame_ind < len(frames_addrs) and addr == frames_addrs[frame_ind]:
                frame = frames[frame_ind]
                print(f"[stack frame {frame}]")
                frame_ind += 1

            identified = identify_address(prog, word_val, cache=cache)
            if identified is None:
                identified = ""
            else:
                identified = f" [{identified}]"
            print(line_format.format(addr, word_val, identified))

        start = end



#!/usr/bin/env drgn
# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""An implementation of lsmod(8) using drgn"""

from drgn.helpers.linux.list import list_for_each_entry

def module_total_size(mod):
    # Since Linux kernel commit ac3b43283923 ("module: replace module_layout
    # with module_memory") (in v6.4), the memory sizes are in the struct
    # module::mem array. Before that, they are in struct module::init_layout
    # and struct module::core_layout.
    try:
        num_types = mod.prog_["MOD_MEM_NUM_TYPES"]
    except KeyError:
        return (mod.init_layout.size + mod.core_layout.size).value_()
    else:
        return sum(
            mod.mem[type].size.value_()
            for type in range(num_types)
        )


print("Module                  Size  Used by")
config_module_unload = prog.type("struct module").has_member("refcnt")
for mod in list_for_each_entry("struct module", prog["modules"].address_of_(), "list"):
    name = mod.name.string_().decode()
    if config_module_unload:
        refcnt = mod.refcnt.counter.value_() - 1
        used_by = [
            use.source.name.string_().decode()
            for use in list_for_each_entry(
                "struct module_use", mod.source_list.address_of_(), "source_list"
            )
        ]
    else:
        refcnt = "-"
        used_by = []

    used = ",".join(used_by)
    if used:
        used = " " + used
    print(f"{name:19} {module_total_size(mod):>8}  {refcnt}{used}")



#!/usr/bin/env drgn
# Copyright (c) SUSE Linux.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Dump /proc/vmstat statistics."""

from drgn.helpers.linux.cpumask import for_each_online_cpu
from drgn.helpers.linux.percpu import per_cpu


def print_event_line(event, counter):
    print(f"{event.name:<36} {counter.value_():>16}")


print(f"{'Event':<36} {'Count':>16}")

# For all of the below, we skip the last enumerator item as it holds the number
# of enumerators.

# 1) vm_zone_stat statistics are there since v4.8.
if "vm_zone_stat" in prog:
    print("VM_ZONE_STAT:")
    vm_zone_stat = prog["vm_zone_stat"]
    for event in prog.type("enum zone_stat_item").enumerators[:-1]:
        print_event_line(event, vm_zone_stat[event.value].counter)
    print()

# 2) vm_node_stat statistics are there since v4.8.
if "vm_node_stat" in prog:
    print("VM_NODE_STAT:")
    vm_node_stat = prog["vm_node_stat"]
    for event in prog.type("enum node_stat_item").enumerators[:-1]:
        print_event_line(event, vm_node_stat[event.value].counter)
    print()

# 3) vm_numa_event statistics are there since v5.14. They are only populated if
# CONFIG_NUMA is enabled.
if "node_subsys" in prog and "vm_numa_event" in prog:
    print("VM_NUMA_EVENT:")
    vm_numa_event = prog["vm_numa_event"]
    for event in prog.type("enum numa_stat_item").enumerators[:-1]:
        print_event_line(event, vm_numa_event[event.value].counter)
    print()

# 4) vm_event_states statistics (uses per-CPU counters)
print("VM_EVENT_STATES:")
vm_event_states = prog["vm_event_states"]
cpulist = list(for_each_online_cpu(prog))
for event in prog.type("enum vm_event_item").enumerators[:-1]:
    count = sum([per_cpu(vm_event_states, cpu).event[event.value] for cpu in cpulist])
    print_event_line(event, count)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Block Layer
-----------

The ``drgn.helpers.linux.block`` module provides helpers for working with the
Linux block layer, including disks (``struct gendisk``) and partitions.

Since Linux v5.11, partitions are represented by ``struct block_device``.
Before that, they were represented by ``struct hd_struct``.
"""

from typing import Iterator

from drgn import Object, Program, cast, container_of
from drgn.helpers.common.format import escape_ascii_string
from drgn.helpers.common.prog import takes_program_or_default
from drgn.helpers.linux.device import MAJOR, MINOR, MKDEV
from drgn.helpers.linux.list import list_for_each_entry

__all__ = (
    "bdev_partno",
    "disk_devt",
    "disk_name",
    "for_each_disk",
    "for_each_partition",
    "part_devt",
    "part_name",
    "print_disks",
    "print_partitions",
)


def disk_devt(disk: Object) -> Object:
    """
    Get a disk's device number.

    :param disk: ``struct gendisk *``
    :return: ``dev_t``
    """
    return Object(disk.prog_, "dev_t", MKDEV(disk.major, disk.first_minor))


def disk_name(disk: Object) -> bytes:
    """
    Get the name of a disk (e.g., ``sda``).

    :param disk: ``struct gendisk *``
    """
    return disk.disk_name.string_()


def _bdev_partno_flags(bdev: Object) -> Object:
    return cast("u8", bdev.__bd_flags.counter)


def _bdev_partno_old(bdev: Object) -> Object:
    return bdev.bd_partno.read_()


def bdev_partno(bdev: Object) -> Object:
    """
    Get the partition number of a block device.

    :param bdev: ``struct block_device *``
    :return: ``u8``
    """
    try:
        impl = bdev.prog_.cache["bdev_partno"]
    except KeyError:
        # Since Linux kernel commit 1116b9fa15c0 ("bdev: infrastructure for
        # flags") (in v6.10), partno is part of the atomic_t __bd_flags member.
        # Before that, it's its own member.
        bdev.prog_.cache["bdev_partno"] = impl = (
            _bdev_partno_flags
            if bdev.prog_.type("struct block_device").has_member("__bd_flags")
            else _bdev_partno_old
        )
    return impl(bdev)


def _class_to_subsys(class_: Object) -> Object:
    # Walk the list of registered classes to find the struct subsys_private
    # matching the given class. Note that before Linux kernel commit
    # 2df418cf4b72 ("driver core: class: remove subsystem private pointer from
    # struct class") (in v6.4), struct subsys_private could also be found in
    # struct class::p, but it's easier to only maintain the newer code path.
    for sp in list_for_each_entry(
        "struct subsys_private",
        class_.prog_["class_kset"].list.address_of_(),
        "subsys.kobj.entry",
    ):
        if sp.member_("class") == class_:
            return sp
    else:
        raise LookupError("block_class subsys_private not found")


def _for_each_block_device(prog: Program) -> Iterator[Object]:
    try:
        devices, class_in_device_private = prog.cache["_for_each_block_device"]
    except KeyError:
        devices = _class_to_subsys(
            prog["block_class"].address_of_()
        ).klist_devices.k_list.address_of_()
        # Linux kernel commit 570d0200123f ("driver core: move
        # device->knode_class to device_private") (in v5.1) moved the list
        # node.
        class_in_device_private = prog.type("struct device_private").has_member(
            "knode_class"
        )
        prog.cache["_for_each_block_device"] = devices, class_in_device_private

    if class_in_device_private:
        for device_private in list_for_each_entry(
            "struct device_private", devices, "knode_class.n_node"
        ):
            yield device_private.device
    else:
        yield from list_for_each_entry("struct device", devices, "knode_class.n_node")


@takes_program_or_default
def for_each_disk(prog: Program) -> Iterator[Object]:
    """
    Iterate over all disks in the system.

    :return: Iterator of ``struct gendisk *`` objects.
    """
    # Before Linux kernel commit 0d02129e76ed ("block: merge struct
    # block_device and struct hd_struct") (in v5.11), partition devices are in
    # struct hd_struct::__dev. After that commit, they are in struct
    # block_device::bd_device. We start by assuming that the kernel has this
    # commit and fall back to the old path if that fails.
    have_bd_device = True
    for device in _for_each_block_device(prog):
        if have_bd_device:
            try:
                bdev = container_of(device, "struct block_device", "bd_device")
            except LookupError:
                have_bd_device = False
            else:
                if not bdev_partno(bdev):
                    yield bdev.bd_disk
                continue
        part = container_of(device, "struct hd_struct", "__dev")
        if part.partno == 0:
            yield container_of(part, "struct gendisk", "part0")


@takes_program_or_default
def print_disks(prog: Program) -> None:
    """Print all of the disks in the system."""
    for disk in for_each_disk(prog):
        major = disk.major.value_()
        minor = disk.first_minor.value_()
        name = escape_ascii_string(disk_name(disk), escape_backslash=True)
        print(f"{major}:{minor} {name} ({disk.type_.type_name()})0x{disk.value_():x}")


def part_devt(part: Object) -> Object:
    """
    Get a partition's device number.

    :param part: ``struct block_device *`` or ``struct hd_struct *`` depending
        on the kernel version.
    :return: ``dev_t``
    """
    try:
        return part.bd_dev
    except AttributeError:
        return part.__dev.devt


def part_name(part: Object) -> bytes:
    """
    Get the name of a partition (e.g., ``sda1``).

    :param part: ``struct block_device *`` or ``struct hd_struct *`` depending
        on the kernel version.
    """
    try:
        bd_device = part.bd_device
    except AttributeError:
        return part.__dev.kobj.name.string_()
    return bd_device.kobj.name.string_()


@takes_program_or_default
def for_each_partition(prog: Program) -> Iterator[Object]:
    """
    Iterate over all partitions in the system.

    :return: Iterator of ``struct block_device *`` or ``struct hd_struct *``
        objects depending on the kernel version.
    """
    # See the comment in for_each_disk().
    have_bd_device = True
    for device in _for_each_block_device(prog):
        if have_bd_device:
            try:
                yield container_of(device, "struct block_device", "bd_device")
                continue
            except LookupError:
                have_bd_device = False
        yield container_of(device, "struct hd_struct", "__dev")


@takes_program_or_default
def print_partitions(prog: Program) -> None:
    """Print all of the partitions in the system."""
    for part in for_each_partition(prog):
        devt = part_devt(part).value_()
        name = escape_ascii_string(part_name(part), escape_backslash=True)
        print(
            f"{MAJOR(devt)}:{MINOR(devt)} {name} ({part.type_.type_name()})0x{part.value_():x}"
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import platform
import re

NORMALIZED_MACHINE_NAME = platform.machine()
if NORMALIZED_MACHINE_NAME.startswith("aarch64") or NORMALIZED_MACHINE_NAME == "arm64":
    NORMALIZED_MACHINE_NAME = "aarch64"
elif NORMALIZED_MACHINE_NAME.startswith("arm") or NORMALIZED_MACHINE_NAME == "sa110":
    NORMALIZED_MACHINE_NAME = "arm"
elif re.fullmatch(r"i.86", NORMALIZED_MACHINE_NAME):
    NORMALIZED_MACHINE_NAME = "i386"
elif NORMALIZED_MACHINE_NAME.startswith("ppc64"):
    NORMALIZED_MACHINE_NAME = "ppc64"
elif NORMALIZED_MACHINE_NAME.startswith("ppc"):
    NORMALIZED_MACHINE_NAME = "ppc"
elif NORMALIZED_MACHINE_NAME == "riscv":
    NORMALIZED_MACHINE_NAME = "riscv32"
elif re.match(r"sh[0-9]", NORMALIZED_MACHINE_NAME):
    NORMALIZED_MACHINE_NAME = "sh"
elif NORMALIZED_MACHINE_NAME == "sun4u":
    NORMALIZED_MACHINE_NAME = "sparc64"


SYS = {
    "aarch64": {
        "bpf": 280,
        "finit_module": 273,
        "kexec_file_load": 294,
        "memfd_create": 279,
        "perf_event_open": 241,
    },
    "alpha": {
        "bpf": 515,
        "finit_module": 507,
        "memfd_create": 512,
        "perf_event_open": 493,
    },
    "arc": {
        "bpf": 280,
        "finit_module": 273,
        "kexec_file_load": 294,
        "memfd_create": 279,
        "perf_event_open": 241,
    },
    "arm": {
        "bpf": 386,
        "finit_module": 379,
        "kexec_file_load": 401,
        "memfd_create": 385,
        "perf_event_open": 364,
    },
    "csky": {
        "bpf": 280,
        "finit_module": 273,
        "kexec_file_load": 294,
        "memfd_create": 279,
        "perf_event_open": 241,
    },
    "hexagon": {
        "bpf": 280,
        "finit_module": 273,
        "kexec_file_load": 294,
        "memfd_create": 279,
        "perf_event_open": 241,
    },
    "i386": {
        "bpf": 357,
        "finit_module": 350,
        "memfd_create": 356,
        "perf_event_open": 336,
    },
    "loongarch": {
        "bpf": 280,
        "finit_module": 273,
        "kexec_file_load": 294,
        "memfd_create": 279,
        "perf_event_open": 241,
    },
    "loongarch64": {
        "bpf": 280,
        "finit_module": 273,
        "kexec_file_load": 294,
        "memfd_create": 279,
        "perf_event_open": 241,
    },
    "m68k": {
        "bpf": 354,
        "finit_module": 348,
        "memfd_create": 353,
        "perf_event_open": 332,
    },
    "microblaze": {
        "bpf": 387,
        "finit_module": 380,
        "memfd_create": 386,
        "perf_event_open": 366,
    },
    # TODO: mips is missing here because I don't know how to distinguish
    # between the o32 and n32 ABIs.
    "mips64": {
        "bpf": 315,
        "finit_module": 307,
        "memfd_create": 314,
        "perf_event_open": 292,
    },
    "nios2": {
        "bpf": 280,
        "finit_module": 273,
        "kexec_file_load": 294,
        "memfd_create": 279,
        "perf_event_open": 241,
    },
    "openrisc": {
        "bpf": 280,
        "finit_module": 273,
        "kexec_file_load": 294,
        "memfd_create": 279,
        "perf_event_open": 241,
    },
    "parisc": {
        "bpf": 341,
        "finit_module": 333,
        "kexec_file_load": 355,
        "memfd_create": 340,
        "perf_event_open": 318,
    },
    "parisc64": {
        "bpf": 341,
        "finit_module": 333,
        "kexec_file_load": 355,
        "memfd_create": 340,
        "perf_event_open": 318,
    },
    "ppc": {
        "bpf": 361,
        "finit_module": 353,
        "memfd_create": 360,
        "perf_event_open": 319,
    },
    "ppc64": {
        "bpf": 361,
        "finit_module": 353,
        "memfd_create": 360,
        "perf_event_open": 319,
    },
    "riscv32": {
        "bpf": 280,
        "finit_module": 273,
        "kexec_file_load": 294,
        "memfd_create": 279,
        "perf_event_open": 241,
    },
    "riscv64": {
        "bpf": 280,
        "finit_module": 273,
        "kexec_file_load": 294,
        "memfd_create": 279,
        "perf_event_open": 241,
    },
    "s390": {
        "bpf": 351,
        "finit_module": 344,
        "kexec_file_load": 381,
        "memfd_create": 350,
        "perf_event_open": 331,
    },
    "s390x": {
        "bpf": 351,
        "finit_module": 344,
        "kexec_file_load": 381,
        "memfd_create": 350,
        "perf_event_open": 331,
    },
    "sh": {
        "bpf": 375,
        "finit_module": 368,
        "memfd_create": 374,
        "perf_event_open": 336,
    },
    "sparc": {
        "bpf": 349,
        "finit_module": 342,
        "memfd_create": 348,
        "perf_event_open": 327,
    },
    "sparc64": {
        "bpf": 349,
        "finit_module": 342,
        "memfd_create": 348,
        "perf_event_open": 327,
    },
    "x86_64": {
        "bpf": 321,
        "finit_module": 313,
        "kexec_file_load": 320,
        "memfd_create": 319,
        "perf_event_open": 298,
    },
    "xtensa": {
        "bpf": 340,
        "finit_module": 332,
        "memfd_create": 339,
        "perf_event_open": 327,
    },
}.get(NORMALIZED_MACHINE_NAME, {})



# Copyright (c) 2023, Oracle and/or its affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Wait Queues
-----------

The ``drgn.helpers.linux.wait`` module provides helpers for working with wait
queues (``wait_queue_head_t`` and ``wait_queue_entry_t``) from
:linux:`include/linux/wait.h`.

.. note::

    Since Linux 4.13, entries in a wait queue have type ``wait_queue_entry_t``.
    Before that, the type was named ``wait_queue_t``.
"""

from typing import Iterator

from drgn import Object, cast
from drgn.helpers.linux.list import list_empty, list_for_each_entry

__all__ = (
    "waitqueue_active",
    "waitqueue_for_each_entry",
    "waitqueue_for_each_task",
)


def _get_wait_queue_head(wq: Object) -> Object:
    # Linux kernel commit 2055da97389a ("sched/wait: Disambiguate
    # wq_entry->task_list and wq_head->task_list naming") (in v4.13) renamed
    # the task_list member to head.
    try:
        return wq.head
    except AttributeError:
        return wq.task_list


def waitqueue_active(wq: Object) -> bool:
    """
    Return whether a wait queue has any waiters.

    :param wq: ``wait_queue_head_t *``
    """
    head = _get_wait_queue_head(wq)
    return not list_empty(head.address_of_())


def waitqueue_for_each_entry(wq: Object) -> Iterator[Object]:
    """
    Iterate over all entries in a wait queue.

    :param wq: ``wait_queue_head_t *``
    :return: Iterator of ``wait_queue_entry_t *`` or ``wait_queue_t *``
        objects depending on the kernel version.
    """
    head_addr = _get_wait_queue_head(wq).address_of_()
    prog = wq.prog_
    # Linux kernel commit ac6424b981bc ("sched/wait: Rename wait_queue_t =>
    # wait_queue_entry_t") (in v4.13) renamed the entry type and commit
    # 2055da97389a ("sched/wait: Disambiguate wq_entry->task_list and
    # wq_head->task_list naming") (in v4.13) renamed .task_list to .entry.
    try:
        wait_queue_entry_type, link = prog.type("wait_queue_entry_t"), "entry"
    except LookupError:
        wait_queue_entry_type, link = prog.type("wait_queue_t"), "task_list"

    return list_for_each_entry(wait_queue_entry_type, head_addr, link)


def waitqueue_for_each_task(wq: Object) -> Iterator[Object]:
    """
    Iterate over all tasks waiting on a wait queue.

    .. warning::

        This comes from ``wait_queue_entry_t::private``, which usually stores a
        task. However, some wait queue entries store a different pointer type,
        in which case this will return garbage.

    :param wq: ``wait_queue_head_t *``
    :return: Iterator of ``struct task_struct *`` objects.
    """
    task_structp_type = wq.prog_.type("struct task_struct *")
    for entry in waitqueue_for_each_entry(wq):
        yield cast(task_structp_type, entry.private)



#!/usr/bin/env drgn
# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""List the paths of all inodes cached in a given filesystem"""

import os
import sys

from drgn.helpers.linux.fs import for_each_mount, inode_path
from drgn.helpers.linux.list import list_for_each_entry

if len(sys.argv) == 1:
    path = "/"
else:
    path = sys.argv[1]

mnt = None
for mnt in for_each_mount(prog, dst=path):
    pass
if mnt is None:
    sys.exit(f"No filesystem mounted at {path}")

sb = mnt.mnt.mnt_sb

for inode in list_for_each_entry(
    "struct inode", sb.s_inodes.address_of_(), "i_sb_list"
):
    try:
        print(os.fsdecode(inode_path(inode)))
    except (TypeError, ValueError):
        continue



#!/usr/bin/env drgn
# Copyright (c) 2023, Oracle and/or its affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

""" Script to dump irq stats using drgn"""

from typing import Iterator
from typing import Tuple

from drgn import NULL
from drgn import Object
from drgn import Program
from drgn.helpers.common.format import escape_ascii_string
from drgn.helpers.linux.cpumask import for_each_present_cpu
from drgn.helpers.linux.cpumask import cpumask_to_cpulist
from drgn.helpers.linux.mapletree import mtree_load
from drgn.helpers.linux.mapletree import mt_for_each
from drgn.helpers.linux.percpu import per_cpu_ptr
from drgn.helpers.linux.radixtree import radix_tree_for_each
from drgn.helpers.linux.radixtree import radix_tree_lookup


def _sparse_irq_supported(prog: Program) -> Tuple[bool, str]:
    try:
        # Since Linux kernel commit 721255b9826b ("genirq: Use a maple
        # tree for interrupt descriptor management") (in v6.5), sparse
        # irq descriptors are stored in a maple tree.
        _ = prog["sparse_irqs"]
        return True, "maple"
    except KeyError:
        # Before that, they are in radix tree.
        try:
            _ = prog["irq_desc_tree"]
            return True, "radix"
        except KeyError:
            return False, None


def _kstat_irqs_cpu(prog: Program, irq: int, cpu: int) -> int:
    desc = irq_to_desc(prog, irq)
    if not desc:
        return 0

    addr = per_cpu_ptr(desc.kstat_irqs, cpu)
    return Object(prog, "int", address=addr).value_()


def irq_in_use(prog: Program, irq: int) -> bool:
    """
    Check if a given irq number is in use or not.
    An irq number is considered to be in use by the kernel, if the kernel
    has allocated a irq descriptor for it. The irq may not yet have any
    registered irq handlers.

    :param prog: drgn program
    :param irq: irq number

    :return: True if irq is in use, False otherwise
    """
    desc = irq_to_desc(prog, irq)
    # An irq number is in use if irq_desc object has been allocated for it
    return bool(desc)


def irq_has_action(prog: Program, irq: int) -> bool:
    """
    Check if a given irq has handler(s) registered or not.

    :param prog: drgn program
    :param irq: irq number

    :return: True if irq has registered handler(s), False otherwise
    """

    desc = irq_to_desc(prog, irq)
    return bool(desc and desc.action)


def for_each_irq(prog: Program) -> Iterator[int]:
    """
    Iterate through all allocated irq numbers.

    :param prog: drgn program

    :return: Iterator of irq numbers
    """
    _, tree_type = _sparse_irq_supported(prog)

    if tree_type == "radix":
        irq_desc_tree = prog["irq_desc_tree"].address_of_()
        for irq, _ in radix_tree_for_each(irq_desc_tree):
            yield irq
    elif tree_type == "maple":
        irq_desc_tree = prog["sparse_irqs"].address_of_()
        for irq, _, _ in mt_for_each(irq_desc_tree):
            yield irq
    else:
        count = len(prog["irq_desc"])
        for irq_num in range(count):
            yield irq_num


def for_each_irq_desc(prog: Program) -> Iterator[Object]:
    """
    Iterate through all allocated irq descriptors.

    :param prog: drgn program

    :return: Iterator of ``struct irq_desc *`` objects.
    """
    _, tree_type = _sparse_irq_supported(prog)
    if tree_type == "radix":
        irq_desc_tree = prog["irq_desc_tree"].address_of_()
        for _, addr in radix_tree_for_each(irq_desc_tree):
            irq_desc = Object(prog, "struct irq_desc", address=addr).address_of_()
            yield irq_desc
    elif tree_type == "maple":
        irq_desc_tree = prog["sparse_irqs"].address_of_()
        for _, _, addr in mt_for_each(irq_desc_tree):
            irq_desc = Object(prog, "struct irq_desc", address=addr).address_of_()
            yield irq_desc
    else:
        count = len(prog["irq_desc"])
        for irq_num in range(count):
            yield (prog["irq_desc"][irq_num]).address_of_()


def irq_name_to_desc(prog: Program, name: str) -> Object:
    """
    Get ``struct irq_desc *`` for irq handler of given name

    :param prog: drgn program
    :param name: name of irq handler

    :return: ``struct irq_desc *`` object if irq descriptor is found.
             NULL otherwise
    """
    for desc in for_each_irq_desc(prog):
        if desc.action and desc.action.name == name:
            return desc

    return NULL(prog, "void *")


def irq_to_desc(prog: Program, irq: int) -> Object:
    """
    Get ``struct irq_desc *`` for given irq number

    :param prog: drgn program
    :param irq: irq number

    :return: ``struct irq_desc *`` object if irq descriptor is found.
             NULL otherwise
    """
    _, tree_type = _sparse_irq_supported(prog)

    if tree_type:
        if tree_type == "radix":
            addr = radix_tree_lookup(prog["irq_desc_tree"].address_of_(), irq)
        else:
            addr = mtree_load(prog["sparse_irqs"].address_of_(), irq)

        if addr:
            return Object(prog, "struct irq_desc", address=addr).address_of_()
        else:
            return NULL(prog, "void *")
    else:
        return (prog["irq_desc"][irq]).address_of_()


def get_irq_affinity(prog: Program, irq: int) -> Object:
    """
    Get ``struct cpumask`` for given irq's cpu affinity

    :param prog: drgn program
    :param irq: irq number

    :return: ``struct cpumask`` object if irq descriptor is found.
             NULL otherwise
    """

    if not irq_in_use(prog, irq):
        print("IRQ not in use so affinity data is not reliable")

    irq_desc = irq_to_desc(prog, irq)

    # if CONFIG_CPUMASK_OFFSTACK is enabled then affinity is an array
    # of cpumask objects otherwise it is pointer to a cpumask object
    if hasattr(irq_desc, "irq_common_data"):
        try:
            _ = len(irq_desc.irq_common_data.affinity)
            addr = irq_desc.irq_common_data.affinity.address_
        except TypeError:
            addr = irq_desc.irq_common_data.affinity.value_()
    elif hasattr(irq_desc, "irq_data"):
        try:
            _ = len(irq_desc.irq_data.affinity)
            addr = irq_desc.irq_data.affinity.address_
        except TypeError:
            addr = irq_desc.irq_data.affinity.value_()
    else:
        return None

    return Object(prog, "struct cpumask", address=addr)


def get_irq_affinity_list(prog: Program, irq: int) -> Object:
    """
    Get affinity of a given cpu.

    :param prog: drgn program
    :param irq: irq number

    :return: range of cpus to which irq is affined to
    """

    affinity = get_irq_affinity(prog, irq)
    if affinity is not None:
        return cpumask_to_cpulist(affinity)
    else:
        return None


def show_irq_num_stats(prog: Program, irq: int) -> None:
    """
    Show stats for a given irq number

    :param prog: drgn program
    :param irq: irq number

    :return: None
    """

    if not irq_in_use(prog, irq):
        print(f"irq: {irq} is not in use")
        return

    if not irq_has_action(prog, irq):
        print(f"irq: {irq} has no handlers registered")
        return
    print_header = True
    total_count = 0
    for cpu in for_each_present_cpu(prog):
        kstat_irqs = _kstat_irqs_cpu(prog, irq, cpu)
        if not kstat_irqs:
            continue
        desc = irq_to_desc(prog, irq)
        name = escape_ascii_string(
            desc.action.name.string_(), escape_backslash=True
        )
        if print_header:
            print(
                f"irq: {irq} name: {name} ({desc.type_.type_name()})0x{desc.value_():x}"
            )
            print_header = False

        total_count += kstat_irqs
        print(f"    CPU: {cpu}  \t count: {kstat_irqs}")

    print(f"    Total: {total_count}")


def show_irq_name_stats(prog: Program, irq_name: str) -> None:
    """
    Show irq stats for irqs whose handler have specified name or
    for irqs whose handler names begin with specified string.

    :param prog: drgn program
    :param irq_name: name or beginning of name of irq handler

    :return: None
    """

    found = False
    for irq in for_each_irq(prog):
        if irq_in_use(prog, irq) and irq_has_action(prog, irq):
            desc = irq_to_desc(prog, irq)
            name = escape_ascii_string(
                desc.action.name.string_(), escape_backslash=True
            )
            if name.startswith(irq_name):
                found = True
                show_irq_num_stats(prog, irq)

    if not found:
        print(
            f"Found no irq with name: {irq_name} or with name starting with: {irq_name}"
        )


def show_irq_stats(prog: Program) -> None:
    """
    Show stats for all irqs.
    :param prog: drgn program

    :return: None
    """
    for irq in for_each_irq(prog):
        if irq_in_use(prog, irq) and irq_has_action(prog, irq):
            show_irq_num_stats(prog, irq)


def show_cpu_irq_num_stats(prog: Program, cpu: int, irq: int) -> None:
    """
    Show irq stats of a cpu for a given irq number

    :param prog: drgn program
    :param cpu: cpu index
    :param irq: irq number

    :return: None
    """

    if not irq_in_use(prog, irq):
        print(f"irq: {irq} is not in use")
        return

    if not irq_has_action(prog, irq):
        print(f"irq: {irq} has no handlers registered")
        return

    print(f"IRQ stats for cpu: {cpu}")
    desc = irq_to_desc(prog, irq)
    name = escape_ascii_string(
        desc.action.name.string_(), escape_backslash=True
    )
    kstat_irqs = _kstat_irqs_cpu(prog, irq, cpu)
    print(
        f"    irq: {irq} name: {name} ({desc.type_.type_name()})0x{desc.value_():x} count: {kstat_irqs}"
    )


def show_cpu_irq_name_stats(prog: Program, cpu: int, irq_name: str) -> None:
    """
    Show irq stats of a cpu for irqs whose handler have specified name or
    for irqs whose handler names begin with specified string.

    :param prog: drgn program
    :param cpu: cpu index
    :param irq_name: name or beginning of name of irq handler

    :return: None
    """

    found = False
    total_irqs_on_cpu = 0
    print(f"IRQ stats for cpu: {cpu}")
    for irq in for_each_irq(prog):
        if irq_in_use(prog, irq) and irq_has_action(prog, irq):
            desc = irq_to_desc(prog, irq)
            name = escape_ascii_string(
                desc.action.name.string_(), escape_backslash=True
            )
            if name.startswith(irq_name):
                found = True
                kstat_irqs = _kstat_irqs_cpu(prog, irq, cpu)
                if not kstat_irqs:
                    continue
                total_irqs_on_cpu += kstat_irqs
                print(
                    f"    irq: {irq} name: {name} ({desc.type_.type_name()})0x{desc.value_():x} count: {kstat_irqs}"
                )

    if not found:
        print(
            f"Found no irq with name: {irq_name} or with name starting with: {irq_name}"
        )
    else:
        print(f"Total: {total_irqs_on_cpu}")


def show_cpu_irq_stats(prog: Program, cpu: int) -> None:
    """
    Show irq stats for specified cpu.

    :param prog: drgn program
    :param cpu: cpu index

    :return: None
    """
    total_irqs_on_cpu = 0
    print(f"IRQ stats for cpu: {cpu}")
    for irq in for_each_irq(prog):
        if irq_in_use(prog, irq) and irq_has_action(prog, irq):
            kstat_irqs = _kstat_irqs_cpu(prog, irq, cpu)
            if not kstat_irqs:
                continue

            desc = irq_to_desc(prog, irq)
            name = escape_ascii_string(
                desc.action.name.string_(), escape_backslash=True
            )
            print(
                f"    irq: {irq} name: {name} ({desc.type_.type_name()})0x{desc.value_():x} count: {kstat_irqs}"
            )
            total_irqs_on_cpu += kstat_irqs

    print(f"Total: {total_irqs_on_cpu}")


def show_each_cpu_irq_stats(prog: Program) -> None:
    """
    Show irq stats for each cpu.

    :param prog: drgn program

    :return: None
    """
    for cpu in for_each_present_cpu(prog):
        show_cpu_irq_stats(prog, cpu)
        print("\n")


def print_irq_affinity(prog: Program, irq: int) -> None:
    """
    Print cpu affinity of specified irq.

    :param prog: drgn program
    :param irq: irq number

    :return: None
    """

    if not irq_in_use(prog, irq):
        print(f"irq: {irq} is not in use")
        return

    if not irq_has_action(prog, irq):
        print(f"irq: {irq} has no handlers registered")
        return

    desc = irq_to_desc(prog, irq)
    name = escape_ascii_string(
        desc.action.name.string_(), escape_backslash=True
    )
    affinity = get_irq_affinity_list(prog, irq)
    print(f"irq: {irq} name: {name} affinity: {affinity}")


def print_irqs_affinities(prog: Program) -> None:
    """
    Print cpu affinities for all irqs in use.

    :param prog: drgn program

    :return: None
    """
    for irq in for_each_irq(prog):
        if irq_in_use(prog, irq) and irq_has_action(prog, irq):
            print_irq_affinity(prog, irq)


def print_all_irqs(prog: Program) -> None:
    """
    Print number, name, ``struct irq_desc *`` and ``struct irqaction *`` for all irqs in use.

    :param prog: drgn program

    :return: None
    """
    for irq in for_each_irq(prog):
        if irq_in_use(prog, irq) and irq_has_action(prog, irq):
            desc = irq_to_desc(prog, irq)
            name = escape_ascii_string(
                desc.action.name.string_(), escape_backslash=True
            )
            print(
                f"irq: {irq} name: {name} ({desc.type_.type_name()})0x{desc.value_():x}  ({desc.action.type_.type_name()})0x{desc.action.value_():x}"
            )


print("###################################################")
print("List of IRQs")
print("###################################################")
print_all_irqs(prog)
print("\n")

print("###################################################")
print("IRQ affinities")
print("###################################################")
print_irqs_affinities(prog)
print("\n")

print("###################################################")
print("IRQ stats")
print("###################################################")
show_irq_stats(prog)
print("\n")

print("###################################################")
print("cpuwise IRQ stats")
print("###################################################")
show_each_cpu_irq_stats(prog)
print("\n")



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Devices
-------

The ``drgn.helpers.linux.device`` module provides helpers for working with
Linux devices, including the kernel encoding of ``dev_t``.
"""

import operator

from drgn import IntegerLike

__all__ = (
    "MAJOR",
    "MINOR",
    "MKDEV",
)


# This hasn't changed since at least v2.6.
_MINORBITS = 20
_MINORMASK = (1 << _MINORBITS) - 1


def MAJOR(dev: IntegerLike) -> int:
    """
    Return the major ID of a kernel ``dev_t``.

    :param dev: ``dev_t`` object or :class:`int`.
    """
    return operator.index(dev) >> _MINORBITS


def MINOR(dev: IntegerLike) -> int:
    """
    Return the minor ID of a kernel ``dev_t``.

    :param dev: ``dev_t`` object or :class:`int`.
    """
    return operator.index(dev) & _MINORMASK


def MKDEV(major: IntegerLike, minor: IntegerLike) -> int:
    """
    Return a kernel ``dev_t`` from the major and minor IDs.

    :param major: Device major ID.
    :param minor: Device minor ID.
    """
    return (operator.index(major) << _MINORBITS) | operator.index(minor)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
CPU Scheduler
-------------

The ``drgn.helpers.linux.sched`` module provides helpers for working with the
Linux CPU scheduler.
"""

from typing import Tuple

from _drgn import (
    _linux_helper_cpu_curr,
    _linux_helper_idle_task,
    _linux_helper_task_cpu as task_cpu,
    _linux_helper_task_thread_info as task_thread_info,
)
from drgn import IntegerLike, Object, Program
from drgn.helpers.common.prog import takes_program_or_default

__all__ = (
    "cpu_curr",
    "idle_task",
    "loadavg",
    "task_cpu",
    "task_state_to_char",
    "task_thread_info",
)

_TASK_NOLOAD = 0x400


@takes_program_or_default
def cpu_curr(prog: Program, cpu: IntegerLike) -> Object:
    """
    Return the task running on the given CPU.

    >>> cpu_curr(7).comm
    (char [16])"python3"

    :param cpu: CPU number.
    :return: ``struct task_struct *``
    """
    return _linux_helper_cpu_curr(prog, cpu)


@takes_program_or_default
def idle_task(prog: Program, cpu: IntegerLike) -> Object:
    """
    Return the idle thread (PID 0, a.k.a swapper) for the given CPU.

    >>> idle_task(1).comm
    (char [16])"swapper/1"

    :param cpu: CPU number.
    :return: ``struct task_struct *``
    """
    return _linux_helper_idle_task(prog, cpu)


def task_state_to_char(task: Object) -> str:
    """
    Get the state of the task as a character (e.g., ``'R'`` for running). See
    `ps(1)
    <http://man7.org/linux/man-pages/man1/ps.1.html#PROCESS_STATE_CODES>`_ for
    a description of the process state codes.

    :param task: ``struct task_struct *``
    """
    prog = task.prog_
    task_state_chars: str
    TASK_REPORT: int
    try:
        task_state_chars, TASK_REPORT, task_state_name = prog.cache[
            "task_state_to_char"
        ]
    except KeyError:
        task_state_array = prog["task_state_array"]
        # Walk through task_state_array backwards looking for the largest state
        # that we know is in TASK_REPORT, then populate the task state mapping.
        chars = None
        for i in range(len(task_state_array) - 1, -1, -1):
            c: int = task_state_array[i][0].value_()
            if chars is None and c in b"RSDTtXZP":
                chars = bytearray(i + 1)
                TASK_REPORT = (1 << i) - 1
            if chars is not None:
                chars[i] = c
        if chars is None:
            raise Exception("could not parse task_state_array")
        task_state_chars = chars.decode("ascii")

        # Since Linux kernel commit 2f064a59a11f ("sched: Change
        # task_struct::state") (in v5.14), the task state is named "__state".
        # Before that, it is named "state".
        try:
            task_state = task.__state
            task_state_name = "__state"
        except AttributeError:
            task_state = task.state
            task_state_name = "state"

        prog.cache["task_state_to_char"] = (
            task_state_chars,
            TASK_REPORT,
            task_state_name,
        )
    else:
        task_state = getattr(task, task_state_name)
    task_state = task_state.value_()
    exit_state = task.exit_state.value_()
    state = (task_state | exit_state) & TASK_REPORT
    char = task_state_chars[state.bit_length()]
    # States beyond TASK_REPORT are special. As of Linux v5.14, TASK_IDLE is
    # the only one; it is defined as TASK_UNINTERRUPTIBLE | TASK_NOLOAD.
    if char == "D" and (task_state & ~state) == _TASK_NOLOAD:
        return "I"
    else:
        return char


@takes_program_or_default
def loadavg(prog: Program) -> Tuple[float, float, float]:
    """
    Return system load averaged over 1, 5 and 15 minutes as
    tuple of three float values.

    >>> loadavg()
    (2.34, 0.442, 1.33)
    """

    avenrun = prog["avenrun"]
    vals = [avenrun[i].value_() / (1 << 11) for i in range(3)]
    return (vals[0], vals[1], vals[2])



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Types
-----

The ``drgn.helpers.common.type`` module provides generic helpers for working
with types in ways that aren't provided by the core drgn library.
"""

import enum
import operator
import typing
from typing import Container, List, Tuple

from drgn import IntegerLike, Type, TypeKind, TypeMember, sizeof

__all__ = (
    "enum_type_to_class",
    "member_at_offset",
)


def enum_type_to_class(
    type: Type, name: str, exclude: Container[str] = (), prefix: str = ""
) -> typing.Type[enum.IntEnum]:
    """
    Get an :class:`enum.IntEnum` class from an enumerated :class:`drgn.Type`.

    :param type: Enumerated type to convert.
    :param name: Name of the ``IntEnum`` type to create.
    :param exclude: Container (e.g., list or set) of enumerator names to
        exclude from the created ``IntEnum``.
    :param prefix: Prefix to strip from the beginning of enumerator names.
    """
    if type.enumerators is None:
        raise TypeError("enum type is incomplete")
    enumerators = [
        (name[len(prefix) :] if name.startswith(prefix) else name, value)
        for (name, value) in type.enumerators
        if name not in exclude
    ]
    return enum.IntEnum(name, enumerators)  # type: ignore  # python/mypy#4865


def member_at_offset(type: Type, offset: IntegerLike) -> str:
    """
    Return the name of the member at an offset in a type.

    This is effectively the opposite of :func:`~drgn.offsetof()`.

    >>> prog.type('struct list_head')
    struct list_head {
            struct list_head *next;
            struct list_head *prev;
    }
    >>> member_at_offset(prog.type('struct list_head'), 0)
    'next'
    >>> member_at_offset(prog.type('struct list_head'), 8)
    'prev'

    This includes nested structures and array elements:

    >>> prog.type('struct sigpending')
    struct sigpending {
            struct list_head list;
            sigset_t signal;
    }
    >>> prog.type('sigset_t')
    typedef struct {
            unsigned long sig[1];
    } sigset_t
    >>> member_at_offset(prog.type('struct sigpending'), 0)
    'list.next'
    >>> member_at_offset(prog.type('struct sigpending'), 8)
    'list.prev'
    >>> member_at_offset(prog.type('struct sigpending'), 16)
    'signal.sig[0]'

    This also includes all possible matches for a union:

    >>> prog.type('union mc_target')
    union mc_target {
            struct folio *folio;
            swp_entry_t ent;
    }
    >>> prog.type('swp_entry_t')
    typedef struct {
            unsigned long val;
    } swp_entry_t
    >>> member_at_offset(prog.type('union mc_target'), 0)
    'folio or ent.val'

    Offsets in the middle of a member are represented:

    >>> member_at_offset(prog.type('struct list_head'), 4)
    'next+0x4'

    Offsets in padding or past the end of the type are also represented:

    >>> prog.type('struct autogroup')
    struct autogroup {
            struct kref kref;
            struct task_group *tg;
            struct rw_semaphore lock;
            unsigned long id;
            int nice;
    }
    >>> member_at_offset(prog.type('struct autogroup'), 4)
    '<padding between kref and tg>'
    >>> member_at_offset(prog.type('struct autogroup'), 70)
    '<padding at end>'
    >>> member_at_offset(prog.type('struct autogroup'), 72)
    '<end>'
    >>> member_at_offset(prog.type('struct autogroup'), 80)
    '<past end>'

    :param type: Type to check.
    :param offset: Offset in bytes.
    :raises TypeError: if *type* is not a structure, union, class, or array
        type (or a typedef of one of those)
    """
    bit_offset = operator.index(offset) * 8

    while type.kind == TypeKind.TYPEDEF:
        type = type.type
    if type.kind not in (
        TypeKind.STRUCT,
        TypeKind.UNION,
        TypeKind.CLASS,
        TypeKind.ARRAY,
    ):
        raise TypeError("must be compound type or array")

    # Chain of member accesses and array subscripts that we've followed.
    chain = []
    # We traverse all union members in a depth-first search. This stack stores
    # the members that still need to be explored, along with the remaining
    # bit_offset from that member and the length of the chain leading to that
    # member.
    stack: List[Tuple[TypeMember, int, int]] = []
    results = []

    # When we've reached the end of a chain, add it to the results and go to
    # the next member in the stack (if any).
    def emit_and_pop_member() -> bool:
        nonlocal type, bit_offset

        if bit_offset:
            if (bit_offset & 7) == 0:
                chain.append("+")
                chain.append(hex(bit_offset // 8))
            else:
                chain.append("+")
                chain.append(str(bit_offset))
                chain.append(" bits")
        results.append("".join(chain))

        if not stack:
            return False

        member, parent_bit_offset, chain_len = stack.pop()

        type = member.type
        bit_offset = parent_bit_offset - member.bit_offset
        del chain[chain_len:]
        if member.name is not None:
            if chain:
                chain.append(".")
            chain.append(member.name)

        return True

    while True:
        if type.kind == TypeKind.TYPEDEF:  # type: ignore[comparison-overlap]  # python/mypy#17096
            type = type.type
        elif type.kind == TypeKind.ARRAY:
            element_bit_size = sizeof(type.type) * 8
            # Treat incomplete arrays as if they have infinite size.
            if type.length is None or bit_offset < type.length * element_bit_size:
                i = bit_offset // element_bit_size
                bit_offset -= i * element_bit_size
                chain.append(f"[{i}]")
                type = type.type
            else:
                if bit_offset == type.length * element_bit_size:
                    chain.append("<end>")
                else:
                    chain.append("<past end>")
                bit_offset = 0
                if not emit_and_pop_member():
                    break
        else:
            members = getattr(type, "members", None)
            if members is None:
                if not emit_and_pop_member():
                    break
                continue

            orig_bit_offset = bit_offset
            orig_chain_len = len(chain)

            # At first, we go forwards through the members. If this is a union,
            # then we go backwards through the rest once we've found a match.
            # This allows us to (1) avoid the stack as an optimization for the
            # common case of structures and (2) return results sorted by
            # declaration order in the source code.
            i = 0
            end = len(members)
            step = 1
            while i != end:
                member = members[i]
                bit_size = member.bit_field_size
                if bit_size is None:
                    try:
                        bit_size = sizeof(member.type) * 8
                    except TypeError:
                        # Ignore incomplete members other than arrays.
                        if member.type.kind != TypeKind.ARRAY:
                            i += step
                            continue
                if (
                    member.bit_offset <= bit_offset
                    # Treat incomplete arrays as if they have infinite size.
                    and (bit_size is None or bit_offset < member.bit_offset + bit_size)
                ):
                    if step == 1:
                        step = -1
                        if type.kind == TypeKind.UNION:
                            i, end = end, i
                        else:
                            # Set i so that we break on the next iteration.
                            i = end + 1

                        type = member.type
                        bit_offset -= member.bit_offset
                        if member.name is not None:
                            if chain:
                                chain.append(".")
                            chain.append(member.name)
                    else:
                        stack.append((member, orig_bit_offset, orig_chain_len))
                i += step
            if step == 1:
                # No matching members.
                bit_size = sizeof(type) * 8
                if bit_offset == bit_size:
                    chain.append("<end>")
                elif bit_offset > bit_size:
                    chain.append("<past end>")
                else:
                    prev_member = None
                    next_member = None
                    for member in members:
                        if member.bit_offset < bit_offset:
                            if (
                                prev_member is None
                                or member.bit_offset > prev_member.bit_offset
                            ):
                                prev_member = member
                        else:
                            if (
                                next_member is None
                                or member.bit_offset < next_member.bit_offset
                            ):
                                next_member = member
                    if chain:
                        chain.append(".")
                    if prev_member and next_member:
                        chain.append("<padding between ")
                        chain.append(
                            "<unnamed>"
                            if prev_member.name is None
                            else prev_member.name
                        )
                        chain.append(" and ")
                        chain.append(
                            "<unnamed>"
                            if next_member.name is None
                            else next_member.name
                        )
                        chain.append(">")
                    elif next_member:
                        chain.append("<padding at beginning>")
                    else:
                        chain.append("<padding at end>")
                bit_offset = 0
                if not emit_and_pop_member():
                    break

    return " or ".join(results)



# Copyright (c) 2024, Oracle and/or its affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Compatibility shim between drgn and the pyrepl/code modules"""

import os
import sys
from typing import Any, Dict

__all__ = ("interact", "readline")

# Python 3.13 introduces a new REPL implemented by the "_pyrepl" internal
# module. It includes features such as colored output and multiline editing.
# Unfortunately, there is no public API exposing these abilities to users, even
# in the "code" module. We'd like to give the best experience possible, so we'll
# detect _pyrepl and try to use it where possible.
try:
    # The official Python interpreter honors this environment variable to
    # disable the new REPL. We do the same, which also gives users an escape
    # hatch if any of the internals we're messing with change.
    if os.environ.get("PYTHON_BASIC_REPL"):
        raise ModuleNotFoundError()

    # Unfortunately, the typeshed library behind mypy explicitly removed type
    # stubs for these modules. This makes sense as they are private APIs, but it
    # means we need to disable mypy checks.
    from _pyrepl import readline  # type: ignore
    from _pyrepl.console import InteractiveColoredConsole  # type: ignore
    from _pyrepl.simple_interact import (  # type: ignore
        run_multiline_interactive_console,
    )

    # This _setup() function clobbers the readline completer, but it is
    # protected so it only runs once. Call it early so that our overridden
    # completer doesn't get clobbered.
    readline._setup({})

    def interact(local: Dict[str, Any], banner: str) -> None:
        console = InteractiveColoredConsole(local)
        print(banner, file=sys.stderr)
        run_multiline_interactive_console(console)

except (ModuleNotFoundError, ImportError, AttributeError):
    import code
    import readline

    def interact(local: Dict[str, Any], banner: str) -> None:
        code.interact(banner=banner, exitmsg="", local=local)



# Copyright (c) 2024 Oracle and/or its affiliates
# SPDX-License-Identifier: LGPL-2.1-or-later
"""
Kallsyms
--------

The ``drgn.helpers.linux.kallsyms`` module contains helpers which allow you to
use the built-in kallsyms symbol table for drgn symbol lookup. Combined with an
alternative type information source, this can enable debugging Linux kernel core
dumps without the corresponding DWARF debuginfo files. Even without type
information, kallsyms can be used to help locate objects, and drgn's low-level
memory reading functions can be used to do basic debugging tasks.
"""
import os
import re
from typing import Dict, List, Tuple

from _drgn import (
    _linux_helper_load_builtin_kallsyms,
    _linux_helper_load_proc_kallsyms as _load_proc_kallsyms,
)
from drgn import (
    Object,
    Program,
    ProgramFlags,
    Symbol,
    SymbolBinding,
    SymbolIndex,
    SymbolKind,
)
from drgn.helpers.linux.module import for_each_module

__all__ = (
    "load_vmlinux_kallsyms",
    "load_module_kallsyms",
)


def _vmcoreinfo_symbols(prog: Program) -> Dict[str, int]:
    vmcoreinfo_data = prog["VMCOREINFO"].string_().decode("ascii")
    vmcoreinfo_symbols = {}
    sym_re = re.compile(r"SYMBOL\(([^)]+)\)=([A-Fa-f0-9]+)")
    for line in vmcoreinfo_data.strip().split("\n"):
        match = sym_re.fullmatch(line)
        if match:
            vmcoreinfo_symbols[match.group(1)] = int(match.group(2), 16)
    return vmcoreinfo_symbols


def _load_builtin_kallsyms(prog: Program) -> SymbolIndex:
    symbol_reqd = [
        "kallsyms_names",
        "kallsyms_token_table",
        "kallsyms_token_index",
        "kallsyms_num_syms",
        "kallsyms_offsets",
        "kallsyms_relative_base",
        "kallsyms_addresses",
        "_stext",
    ]
    symbols = _vmcoreinfo_symbols(prog)
    args = []
    for sym in symbol_reqd:
        args.append(symbols.get(sym, 0))
    return _linux_helper_load_builtin_kallsyms(prog, *args)


def load_vmlinux_kallsyms(prog: Program) -> SymbolIndex:
    """
    Create a kallsyms index for vmlinux

    This function loads the kallsyms for the core kernel and returns a symbol
    index. This function does not require that any debuginfo is loaded for the
    kernel: it either relies on ``/proc/kallsyms`` (which requires running drgn
    as root) or it parses internal data structures using information found from
    the VMCOREINFO note (which requires Linux 6.0 or later, or a backport of
    commit ``f09bddbd86619 ("vmcoreinfo: add kallsyms_num_syms symbol")`` and
    its dependencies).

    :returns: a symbol index containing kallsyms for the core kernel (vmlinux)
    """
    if prog.flags & ProgramFlags.IS_LIVE and os.geteuid() == 0:
        return _load_proc_kallsyms()
    else:
        return _load_builtin_kallsyms(prog)


def _nm_type_to_binding_kind(code: str) -> Tuple[SymbolBinding, SymbolKind]:
    binding = SymbolBinding.UNKNOWN
    kind = SymbolKind.UNKNOWN
    if code == "v":
        binding = SymbolBinding.WEAK
        kind = SymbolKind.OBJECT
    elif code == "w":
        binding = SymbolBinding.WEAK
    elif code in "tT":
        kind = SymbolKind.FUNC
    elif code.lower() in "srbgncd":
        kind = SymbolKind.OBJECT
    if binding == SymbolBinding.UNKNOWN and code.isupper():
        binding = SymbolBinding.GLOBAL
    return binding, kind


def _st_info_to_binding_kind(info: int) -> Tuple[SymbolBinding, SymbolKind]:
    binding_int = info >> 4
    STB_WEAK = 2
    STB_GNU_UNIQUE = 10
    if binding_int <= STB_WEAK or binding_int == STB_GNU_UNIQUE:
        binding = SymbolBinding(binding_int + 1)
    else:
        binding = SymbolBinding.UNKNOWN
    type_ = info & 0xF
    STT_TLS = 6
    STT_GNU_IFUNC = 10
    if type_ <= STT_TLS or type_ == STT_GNU_IFUNC:
        kind = SymbolKind(type_)
    else:
        kind = SymbolKind.UNKNOWN
    return binding, kind


def _elf_sym_to_symbol(name: str, obj: Object, has_typetab: bool) -> Symbol:
    # Linux likes to have the nm(1) character code for its symbols, which it
    # refers to as the symbol's "type" (this is of course distinct from the ELF
    # notion of a symbol type, let alone what drgn considers a "type"...).
    #
    # Prior to 5439c985c5a8 ("module: Overwrite st_size instead of st_info"),
    # merged in v5.0, the kernel simply overwrote the "st_info" field with a
    # single-character code that represents the nm(1) character code for that
    # symbol. However, starting with that commit, it was switched to overwrite
    # the "st_size" field instead! This was thankfully fixed in v5.2 with
    # 1c7651f43777 ("kallsyms: store type information in its own array").
    #
    # Unfortunately, this leaves us with three possibilities:
    # 1. Pre-v5.0: interpret the "st_info" as a character from nm(1) and try to
    #    infer the kind and bindings.
    # 2. 5.0-5.2: interpret the "st_info" as normal, but ignore the "st_size"
    #    field since it is bogus.
    # 3. 5.2+: both fields are valid, and the nm(1) code is stored in "typetab".
    #
    # Case 3 can be determined easily by the presence of "typetab" in "struct
    # mod_kallsyms". However, cases 1 & 2 are indistinguishable. For our
    # purposes, it makes more sense to fall back to case 1. After all, neither
    # 5.0 or 5.1 were LTS kernels, nor are they actively used by any major
    # distro. We have no way to deal with 5.0 or 5.1, whereas we can make some
    # informed guesses for pre-5.0 based on the nm(1) code.
    if has_typetab:
        binding, kind = _st_info_to_binding_kind(obj.st_info.value_())
    else:
        binding, kind = _nm_type_to_binding_kind(chr(obj.st_info.value_()))
    return Symbol(
        name,
        obj.st_value.value_(),
        obj.st_size.value_(),
        binding,
        kind,
    )


def _module_kallsyms(module: Object) -> List[Symbol]:
    try:
        ks = module.kallsyms
    except AttributeError:
        # Prior to 8244062ef1e54 ("modules: fix longstanding /proc/kallsyms vs
        # module insertion race."), the kallsyms variables were stored directly
        # on the module object. This commit was introduced in 4.5, but was
        # backported to some stable kernels too. Fall back to the module object
        # in cases where kallsyms field isn't available.
        ks = module

    prog = module.prog_
    num_symtab = ks.num_symtab.value_()
    try:
        ks.member_("typetab")
        has_typetab = True
    except LookupError:
        has_typetab = False

    # The symtab field is a pointer, but it points at an array of Elf_Sym
    # objects. Indexing it requires drgn to do pointer arithmetic and issue a
    # lot of very small /proc/kcore reads, which can be a real performance
    # issue. So convert it into an object representing a correctly-sized array,
    # and then read that object all at once. This does one /proc/kcore read,
    # which is a major improvement!
    symtab = Object(
        prog,
        type=prog.array_type(ks.symtab.type_.type, num_symtab),
        address=ks.symtab.value_(),
    ).read_()

    # The strtab is similarly a pointer into a contigous array of strings packed
    # next to each other. Reading individual strings from /proc/kcore can be
    # quite slow. So read the entire array of bytes into a Python bytes value,
    # and we'll extract the individual symbol strings from there.
    last_string_start = symtab[num_symtab - 1].st_name.value_()
    last_string_len = len(ks.strtab[last_string_start].address_of_().string_()) + 1
    strtab = prog.read(ks.strtab.value_(), last_string_start + last_string_len)
    syms = []
    for i in range(ks.num_symtab.value_()):
        elfsym = symtab[i]
        if not elfsym.st_name:
            continue
        str_index = elfsym.st_name.value_()
        nul_byte = strtab.find(b"\x00", str_index)
        name = strtab[str_index:nul_byte].decode("ascii")
        syms.append(_elf_sym_to_symbol(name, elfsym, has_typetab))
    return syms


def load_module_kallsyms(prog: Program) -> SymbolIndex:
    """
    Return a symbol index containing all module symbols from kallsyms

    For kernels built with ``CONFIG_KALLSYMS``, loaded kernel modules contain
    an ELF symbol table in kernel memory. This function can parse those data
    structures and create a symbol index usable by drgn. However, it requires
    that you already have debuginfo for the vmlinux image.

    :returns: a symbol index containing all symbols from module kallsyms
    """
    all_symbols = []
    for module in for_each_module(prog):
        all_symbols.extend(_module_kallsyms(module))
    return SymbolIndex(all_symbols)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Experimental
------------

The ``drgn.helpers.experimental`` package contains experimental helpers with no
stability guarantees. They may change, move to another package, or be removed.
They are not automatically imported by the CLI.
"""



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Memory Management
-----------------

The ``drgn.helpers.linux.mm`` module provides helpers for working with the
Linux memory management (MM) subsystem.

.. _virtual address translation failures:

Helpers that translate virtual addresses or read virtual memory may fail for
multiple reasons:

1. If the address is invalid.
2. If the address is swapped or paged out.
3. If the address is in `high memory
   <https://docs.kernel.org/mm/highmem.html>`_. High memory is only used for
   userspace memory by 32-bit systems with a lot of physical memory, and only
   if ``CONFIG_HIGHMEM`` is enabled.

   3a. If the page table is in high memory. This is only possible if
   ``CONFIG_HIGHPTE`` is enabled.
"""

import operator
import re
from typing import Callable, Iterator, List, Optional

from _drgn import (
    _linux_helper_direct_mapping_offset,
    _linux_helper_follow_phys,
    _linux_helper_read_vm,
)
from drgn import NULL, IntegerLike, Object, ObjectAbsentError, Program, cast
from drgn.helpers.common.format import decode_enum_type_flags
from drgn.helpers.common.prog import takes_program_or_default
from drgn.helpers.linux.list import list_for_each_entry
from drgn.helpers.linux.mapletree import mt_for_each, mtree_load
from drgn.helpers.linux.rbtree import rb_find

__all__ = (
    "PFN_PHYS",
    "PHYS_PFN",
    "PageCompound",
    "PageHead",
    "PageSlab",
    "PageTail",
    "access_process_vm",
    "access_remote_vm",
    "cmdline",
    "compound_head",
    "compound_nr",
    "compound_order",
    "decode_page_flags",
    "environ",
    "find_vmap_area",
    "follow_page",
    "follow_pfn",
    "follow_phys",
    "for_each_page",
    "for_each_vma",
    "for_each_vmap_area",
    "page_size",
    "page_to_pfn",
    "page_to_phys",
    "page_to_virt",
    "pfn_to_page",
    "pfn_to_virt",
    "phys_to_page",
    "phys_to_virt",
    "totalram_pages",
    "virt_to_page",
    "virt_to_pfn",
    "virt_to_phys",
    "vma_find",
    "vmalloc_to_page",
    "vmalloc_to_pfn",
    # Generated by scripts/generate_page_flag_getters.py.
    "PageActive",
    "PageChecked",
    "PageDirty",
    "PageDoubleMap",
    "PageError",
    "PageForeign",
    "PageHWPoison",
    "PageHasHWPoisoned",
    "PageIdle",
    "PageIsolated",
    "PageLRU",
    "PageLocked",
    "PageMappedToDisk",
    "PageMlocked",
    "PageOwnerPriv1",
    "PagePinned",
    "PagePrivate",
    "PagePrivate2",
    "PageReadahead",
    "PageReclaim",
    "PageReferenced",
    "PageReported",
    "PageReserved",
    "PageSavePinned",
    "PageSkipKASanPoison",
    "PageSlobFree",
    "PageSwapBacked",
    "PageUncached",
    "PageUnevictable",
    "PageUptodate",
    "PageVmemmapSelfHosted",
    "PageWaiters",
    "PageWorkingset",
    "PageWriteback",
    "PageXenRemapped",
    "PageYoung",
)


def PageActive(page: Object) -> bool:
    """
    Return whether the ``PG_active`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_active"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PageChecked(page: Object) -> bool:
    """
    Return whether the ``PG_checked`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_checked"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PageDirty(page: Object) -> bool:
    """
    Return whether the ``PG_dirty`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_dirty"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PageDoubleMap(page: Object) -> bool:
    """
    Return whether the ``PG_double_map`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_double_map"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PageError(page: Object) -> bool:
    """
    Return whether the ``PG_error`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_error"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PageForeign(page: Object) -> bool:
    """
    Return whether the ``PG_foreign`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_foreign"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PageHWPoison(page: Object) -> bool:
    """
    Return whether the ``PG_hwpoison`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_hwpoison"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PageHasHWPoisoned(page: Object) -> bool:
    """
    Return whether the ``PG_has_hwpoisoned`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_has_hwpoisoned"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PageIdle(page: Object) -> bool:
    """
    Return whether the ``PG_idle`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_idle"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PageIsolated(page: Object) -> bool:
    """
    Return whether the ``PG_isolated`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_isolated"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PageLRU(page: Object) -> bool:
    """
    Return whether the ``PG_lru`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_lru"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PageLocked(page: Object) -> bool:
    """
    Return whether the ``PG_locked`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_locked"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PageMappedToDisk(page: Object) -> bool:
    """
    Return whether the ``PG_mappedtodisk`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_mappedtodisk"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PageMlocked(page: Object) -> bool:
    """
    Return whether the ``PG_mlocked`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_mlocked"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PageOwnerPriv1(page: Object) -> bool:
    """
    Return whether the ``PG_owner_priv_1`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_owner_priv_1"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PagePinned(page: Object) -> bool:
    """
    Return whether the ``PG_pinned`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_pinned"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PagePrivate(page: Object) -> bool:
    """
    Return whether the ``PG_private`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_private"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PagePrivate2(page: Object) -> bool:
    """
    Return whether the ``PG_private_2`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_private_2"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PageReadahead(page: Object) -> bool:
    """
    Return whether the ``PG_readahead`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_readahead"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PageReclaim(page: Object) -> bool:
    """
    Return whether the ``PG_reclaim`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_reclaim"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PageReferenced(page: Object) -> bool:
    """
    Return whether the ``PG_referenced`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_referenced"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PageReported(page: Object) -> bool:
    """
    Return whether the ``PG_reported`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_reported"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PageReserved(page: Object) -> bool:
    """
    Return whether the ``PG_reserved`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_reserved"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PageSavePinned(page: Object) -> bool:
    """
    Return whether the ``PG_savepinned`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_savepinned"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PageSkipKASanPoison(page: Object) -> bool:
    """
    Return whether the ``PG_skip_kasan_poison`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_skip_kasan_poison"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PageSlobFree(page: Object) -> bool:
    """
    Return whether the ``PG_slob_free`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_slob_free"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PageSwapBacked(page: Object) -> bool:
    """
    Return whether the ``PG_swapbacked`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_swapbacked"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PageUncached(page: Object) -> bool:
    """
    Return whether the ``PG_uncached`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_uncached"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PageUnevictable(page: Object) -> bool:
    """
    Return whether the ``PG_unevictable`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_unevictable"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PageUptodate(page: Object) -> bool:
    """
    Return whether the ``PG_uptodate`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_uptodate"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PageVmemmapSelfHosted(page: Object) -> bool:
    """
    Return whether the ``PG_vmemmap_self_hosted`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_vmemmap_self_hosted"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PageWaiters(page: Object) -> bool:
    """
    Return whether the ``PG_waiters`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_waiters"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PageWorkingset(page: Object) -> bool:
    """
    Return whether the ``PG_workingset`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_workingset"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PageWriteback(page: Object) -> bool:
    """
    Return whether the ``PG_writeback`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_writeback"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PageXenRemapped(page: Object) -> bool:
    """
    Return whether the ``PG_xen_remapped`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_xen_remapped"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


def PageYoung(page: Object) -> bool:
    """
    Return whether the ``PG_young`` flag is set on a page.

    :param page: ``struct page *``
    """
    try:
        flag = page.prog_["PG_young"]
    except KeyError:
        return False
    return bool(page.flags & (1 << flag))


# End generated by scripts/generate_page_flag_getters.py.


def _get_PageSlab_impl(prog: Program) -> Callable[[Object], bool]:
    # Since Linux kernel commit 46df8e73a4a3 ("mm: free up PG_slab") (in
    # v6.10), slab pages are identified by a page type, which is indicated by a
    # mapcount value matching a value in VMCOREINFO. Before that, they are
    # indicated by a page flag.
    try:
        return prog.cache["PageSlab"]
    except KeyError:
        pass
    vmcoreinfo = prog["VMCOREINFO"].string_()
    match = re.search(
        rb"^NUMBER\(PAGE_SLAB_MAPCOUNT_VALUE\)=(-?[0-9]+)$", vmcoreinfo, flags=re.M
    )
    if match:
        PAGE_SLAB_MAPCOUNT_VALUE = int(match.group(1))

        def PageSlab(page: Object) -> bool:
            return page._mapcount.counter.value_() == PAGE_SLAB_MAPCOUNT_VALUE

    else:
        mask = 1 << prog["PG_slab"]

        def PageSlab(page: Object) -> bool:
            return bool(page.flags & mask)

    prog.cache["PageSlab"] = PageSlab
    return PageSlab


def PageSlab(page: Object) -> bool:
    """
    Return whether a page belongs to the slab allocator.

    :param page: ``struct page *``
    """
    return _get_PageSlab_impl(page.prog_)(page)


def PageCompound(page: Object) -> bool:
    """
    Return whether a page is part of a `compound page
    <https://lwn.net/Articles/619514/>`_.

    :param page: ``struct page *``
    """
    page = page.read_()
    # Since Linux kernel commit 1d798ca3f164 ("mm: make compound_head()
    # robust") (in v4.4), PG_head is always defined, and a tail page has the
    # least significant bit of compound_head set. Before that, there is no
    # compound_head (and no fake head pages). Instead, if
    # CONFIG_PAGEFLAGS_EXTENDED=y, then PG_head and PG_tail are defined.
    # Otherwise, there is only PG_compound, and PG_reclaim is set for tail
    # pages and clear for head pages.
    try:
        PG_head = page.prog_["PG_head"]
    except KeyError:
        return bool(page.flags & (1 << page.prog_["PG_compound"]))
    else:
        flags = page.flags.read_()
        if flags & (1 << PG_head):
            return True
        try:
            return bool(page.compound_head.read_() & 1)
        except AttributeError:
            return bool(flags & (1 << page.prog_["PG_tail"]))


# HugeTLB Vmemmap Optimization (HVO) creates "fake" head pages that are
# actually tail pages. See Linux kernel commit e7d324850bfc ("mm: hugetlb: free
# the 2nd vmemmap page associated with each HugeTLB page") (in v5.18) and
# https://www.kernel.org/doc/html/latest/mm/vmemmap_dedup.html.
def _page_is_fake_head(page: Object) -> bool:
    head = page[1].compound_head.value_()
    return bool(head & 1) and (head - 1) != page.value_()


def PageHead(page: Object) -> bool:
    """
    Return whether a page is a head page in a `compound page`_.

    :param page: ``struct page *``
    """
    page = page.read_()
    # See PageCompound() re: Linux kernel commit 1d798ca3f164 ("mm: make
    # compound_head() robust") (in v4.4).
    try:
        PG_head = page.prog_["PG_head"]
    except KeyError:
        PG_compound = page.prog_["PG_compound"]
        PG_head_mask = 1 << PG_compound
        PG_head_tail_mask = PG_head_mask | (1 << page.prog_["PG_reclaim"])
        return (page.flags & PG_head_tail_mask) == PG_head_mask
    else:
        if not (page.flags & (1 << PG_head)):
            return False
        try:
            return not _page_is_fake_head(page)
        except AttributeError:
            return True


def PageTail(page: Object) -> bool:
    """
    Return whether a page is a tail page in a `compound page`_.

    :param page: ``struct page *``
    """
    page = page.read_()
    # See PageCompound() re: Linux kernel commit 1d798ca3f164 ("mm: make
    # compound_head() robust") (in v4.4).
    try:
        if page.compound_head.value_() & 1:
            return True
    except AttributeError:
        try:
            PG_tail = page.prog_["PG_tail"]
        except KeyError:
            PG_head_tail_mask = (1 << page.prog_["PG_compound"]) | (
                1 << page.prog_["PG_reclaim"]
            )
            return (page.flags & PG_head_tail_mask) == PG_head_tail_mask
        else:
            return bool(page.flags & (1 << PG_tail))
    if page.flags & (1 << page.prog_["PG_head"]):
        return _page_is_fake_head(page)
    return False


def compound_head(page: Object) -> Object:
    """
    Get the head page associated with a page.

    If *page* is a tail page, this returns the head page of the `compound
    page`_ it belongs to. Otherwise, it returns *page*.

    :param page: ``struct page *``
    :return: ``struct page *``
    """
    page = page.read_()
    try:
        head = page.compound_head.read_()
    except AttributeError:
        # Before Linux kernel commit 1d798ca3f164 ("mm: make compound_head()
        # robust") (in v4.4), the head page is in page->first_page, and there
        # are no fake head pages.
        return page.first_page.read_() if PageTail(page) else page
    if head & 1:
        return cast(page.type_, head - 1)
    # Handle fake head pages (see _page_is_fake_head()).
    if page.flags & (1 << page.prog_["PG_head"]):
        head = page[1].compound_head.read_()
        if head & 1:
            return cast(page.type_, head - 1)
    return page


def compound_order(page: Object) -> Object:
    """
    Return the allocation order of a potentially `compound page`_.

    :param page: ``struct page *``
    :return: ``unsigned int``
    """
    prog = page.prog_

    if not PageHead(page):
        return Object(prog, "unsigned int", 0)

    # Since Linux kernel commit ebc1baf5c9b4 ("mm: free up a word in the first
    # tail page") (in v6.6), the compound order is in the low byte of struct
    # folio::_flags_1 (from_folio = 2). Between that and Linux kernel commit
    # Linux kernel commit 379708ffde1b ("mm: add the first tail page to struct
    # folio") (in v6.1), the compound order is in struct folio::_folio_order
    # (from_folio = 1). Before Linux kernel commit 1c5509be58f6 ("mm: remove
    # 'First tail page' members from struct page") (in v6.3), the compound
    # order is in struct page::compound_order of the first tail page
    # (from_folio = 0).
    try:
        from_folio = prog.cache["compound_order_from_folio"]
    except KeyError:
        from_folio = 0
        try:
            struct_folio = prog.type("struct folio")
        except LookupError:
            pass
        else:
            if struct_folio.has_member("_folio_order"):
                from_folio = 1
            elif struct_folio.has_member("_flags_1"):
                from_folio = 2
        prog.cache["compound_order_from_folio"] = from_folio
    if from_folio == 2:
        return cast("unsigned int", cast("struct folio *", page)._flags_1 & 0xFF)
    elif from_folio == 1:
        return cast("unsigned int", cast("struct folio *", page)._folio_order)
    else:
        return cast("unsigned int", page[1].compound_order)


def compound_nr(page: Object) -> Object:
    """
    Return the number of pages in a potentially `compound page`_.

    :param page: ``struct page *``
    :return: ``unsigned long``
    """
    return Object(page.prog_, "unsigned long", 1) << compound_order(page)


def page_size(page: Object) -> Object:
    """
    Return the number of bytes in a potentially `compound page`_.

    :param page: ``struct page *``
    :return: ``unsigned long``
    """
    return page.prog_["PAGE_SIZE"] << compound_order(page)


def decode_page_flags(page: Object) -> str:
    """
    Get a human-readable representation of the flags set on a page.

    >>> decode_page_flags(page)
    'PG_uptodate|PG_dirty|PG_lru|PG_reclaim|PG_swapbacked|PG_readahead|PG_savepinned|PG_isolated|PG_reported'

    :param page: ``struct page *``
    """
    NR_PAGEFLAGS = page.prog_["__NR_PAGEFLAGS"]
    PAGEFLAGS_MASK = (1 << NR_PAGEFLAGS.value_()) - 1
    return decode_enum_type_flags(
        page.flags.value_() & PAGEFLAGS_MASK, NR_PAGEFLAGS.type_
    )


# Get the struct page * for PFN 0.
def _page0(prog: Program) -> Object:
    try:
        return prog.cache["page0"]
    except KeyError:
        pass
    try:
        # With CONFIG_SPARSEMEM_VMEMMAP=y, page 0 is vmemmap.
        page0 = prog["vmemmap"]
    except KeyError:
        contig_page_data = prog["contig_page_data"]
        # With CONFIG_FLATMEM=y, page 0 is mem_map - ARCH_PFN_OFFSET, but we
        # can't determine ARCH_PFN_OFFSET easily. Alternatively,
        # contig_page_data.node_mem_map is the struct page * for
        # contig_page_data.node_start_pfn, therefore page 0 is:
        page0 = contig_page_data.node_mem_map - contig_page_data.node_start_pfn
    # The struct page array is not contiguous for CONFIG_SPARSEMEM=y with
    # CONFIG_SPARSEMEM_VMEMMAP=n or CONFIG_DISCONTIGMEM=y, so those are not
    # supported yet.
    prog.cache["page0"] = page0
    return page0


@takes_program_or_default
def for_each_page(prog: Program) -> Iterator[Object]:
    """
    Iterate over every ``struct page *`` from the minimum to the maximum page.

    .. note::

        This may include offline pages which don't have a valid ``struct
        page``. Wrap accesses in a ``try`` ... ``except``
        :class:`drgn.FaultError`:

        >>> for page in for_each_page():
        ...     try:
        ...         if PageLRU(page):
        ...             print(hex(page))
        ...     except drgn.FaultError:
        ...         continue
        0xfffffb4a000c0000
        0xfffffb4a000c0040
        ...

        This may be fixed in the future.

    :return: Iterator of ``struct page *`` objects.
    """
    page0 = _page0(prog)
    for i in range(prog["min_low_pfn"], prog["max_pfn"]):
        yield page0 + i


@takes_program_or_default
def PFN_PHYS(prog: Program, pfn: IntegerLike) -> Object:
    """
    Get the physical address of a page frame number (PFN).

    :param pfn: ``unsigned long``
    :return: ``phys_addr_t``
    """
    return Object(prog, "phys_addr_t", pfn) << prog["PAGE_SHIFT"]


@takes_program_or_default
def PHYS_PFN(prog: Program, addr: IntegerLike) -> Object:
    """
    Get the page frame number (PFN) of a physical address.

    :param addr: ``phys_addr_t``
    :return: ``unsigned long``
    """
    return Object(prog, "unsigned long", addr) >> prog["PAGE_SHIFT"]


def page_to_pfn(page: Object) -> Object:
    """
    Get the page frame number (PFN) of a page.

    :param page: ``struct page *``
    :return: ``unsigned long``
    """
    return cast("unsigned long", page - _page0(page.prog_))


def page_to_phys(page: Object) -> Object:
    """
    Get the physical address of a page.

    :param page: ``struct page *``
    :return: ``phys_addr_t``
    """
    return PFN_PHYS(page_to_pfn(page))


def page_to_virt(page: Object) -> Object:
    """
    Get the directly mapped virtual address of a page.

    :param page: ``struct page *``
    :return: ``void *``
    """
    return pfn_to_virt(page_to_pfn(page))


@takes_program_or_default
def pfn_to_page(prog: Program, pfn: IntegerLike) -> Object:
    """
    Get the page with a page frame number (PFN).

    :param pfn: ``unsigned long``
    :return: ``struct page *``
    """
    return _page0(prog) + pfn


@takes_program_or_default
def pfn_to_virt(prog: Program, pfn: IntegerLike) -> Object:
    """
    Get the directly mapped virtual address of a page frame number (PFN).

    :param pfn: ``unsigned long``
    :return: ``void *``
    """
    return phys_to_virt(PFN_PHYS(prog, pfn))


@takes_program_or_default
def phys_to_page(prog: Program, addr: IntegerLike) -> Object:
    """
    Get the page containing a physical address.

    :param addr: ``phys_addr_t``
    :return: ``struct page *``
    """
    return pfn_to_page(PHYS_PFN(prog, addr))


@takes_program_or_default
def phys_to_virt(prog: Program, addr: IntegerLike) -> Object:
    """
    Get the directly mapped virtual address of a physical address.

    :param addr: ``phys_addr_t``
    :return: ``void *``
    """
    return Object(
        prog, "void *", operator.index(addr) + _linux_helper_direct_mapping_offset(prog)
    )


@takes_program_or_default
def virt_to_page(prog: Program, addr: IntegerLike) -> Object:
    """
    Get the page containing a directly mapped virtual address.

    .. _mm-helpers-direct-map:

    .. note::

        This only works for virtual addresses from the "direct map". This
        includes address from:

        * kmalloc
        * Slab allocator
        * Page allocator

        But not:

        * vmalloc
        * vmap
        * ioremap
        * Symbols (function pointers, global variables)

        For vmalloc or vmap addresses, use :func:`vmalloc_to_page(addr)
        <vmalloc_to_page>`. For arbitrary kernel addresses, use
        :func:`follow_page(prog["init_mm"].address_of_(), addr) <follow_page>`.

    :param addr: ``void *``
    :return: ``struct page *``
    """
    return pfn_to_page(virt_to_pfn(prog, addr))


@takes_program_or_default
def virt_to_pfn(prog: Program, addr: IntegerLike) -> Object:
    """
    Get the page frame number (PFN) of a directly mapped virtual address.

    .. note::

        This only works for virtual addresses from the :ref:`"direct map"
        <mm-helpers-direct-map>`. For vmalloc or vmap addresses, use
        :func:`vmalloc_to_pfn(addr) <vmalloc_to_pfn>`. For arbitrary kernel
        addresses, use :func:`follow_pfn(prog["init_mm"].address_of_(), addr)
        <follow_pfn>`.

    :param addr: ``void *``
    :return: ``unsigned long``
    """
    return PHYS_PFN(virt_to_phys(prog, addr))


@takes_program_or_default
def virt_to_phys(prog: Program, addr: IntegerLike) -> Object:
    """
    Get the physical address of a directly mapped virtual address.

    .. note::

        This only works for virtual addresses from the :ref:`"direct map"
        <mm-helpers-direct-map>`. For arbitrary kernel addresses, use
        :func:`follow_phys(prog["init_mm"].address_of_(), addr) <follow_phys>`.

    :param addr: ``void *``
    :return: ``phys_addr_t``
    """
    return Object(
        prog,
        "unsigned long",
        operator.index(addr) - _linux_helper_direct_mapping_offset(prog),
    )


def follow_page(mm: Object, addr: IntegerLike) -> Object:
    """
    Get the page that a virtual address maps to in a virtual address space.

    >>> task = find_task(113)
    >>> follow_page(task.mm, 0x7fffbbb6d4d0)
    *(struct page *)0xffffbe4bc0337b80 = {
        ...
    }

    :param mm: ``struct mm_struct *``
    :param addr: ``void *``
    :return: ``struct page *``
    :raises FaultError: if the virtual address
        :ref:`cannot be translated <virtual address translation failures>`
    :raises NotImplementedError: if virtual address translation is :ref:`not
        supported <architecture support matrix>` for this architecture yet
    """
    return phys_to_page(follow_phys(mm, addr))


def follow_pfn(mm: Object, addr: IntegerLike) -> Object:
    """
    Get the page frame number (PFN) that a virtual address maps to in a virtual
    address space.

    >>> task = find_task(113)
    >>> follow_pfn(task.mm, 0x7fffbbb6d4d0)
    (unsigned long)52718

    :param mm: ``struct mm_struct *``
    :param addr: ``void *``
    :return: ``unsigned long``
    :raises FaultError: if the virtual address
        :ref:`cannot be translated <virtual address translation failures>`
    :raises NotImplementedError: if virtual address translation is :ref:`not
        supported <architecture support matrix>` for this architecture yet
    """
    return PHYS_PFN(follow_phys(mm, addr))


def follow_phys(mm: Object, addr: IntegerLike) -> Object:
    """
    Get the physical address that a virtual address maps to in a virtual
    address space.

    >>> task = find_task(113)
    >>> follow_phys(task.mm, 0x7fffbbb6d4d0)
    (phys_addr_t)215934160

    :param mm: ``struct mm_struct *``
    :param addr: ``void *``
    :return: ``phys_addr_t``
    :raises FaultError: if the virtual address
        :ref:`cannot be translated <virtual address translation failures>`
    :raises NotImplementedError: if virtual address translation is :ref:`not
        supported <architecture support matrix>` for this architecture yet
    """
    prog = mm.prog_
    return Object(prog, "phys_addr_t", _linux_helper_follow_phys(prog, mm.pgd, addr))


@takes_program_or_default
def vmalloc_to_page(prog: Program, addr: IntegerLike) -> Object:
    """
    Get the page containing a vmalloc or vmap address.

    >>> task = find_task(113)
    >>> vmalloc_to_page(task.stack)
    *(struct page *)0xffffbe4bc00a2200 = {
        ...
    }

    :param addr: ``void *``
    :return: ``struct page *``
    """
    return follow_page(prog["init_mm"].address_of_(), addr)


@takes_program_or_default
def vmalloc_to_pfn(prog: Program, addr: IntegerLike) -> Object:
    """
    Get the page frame number (PFN) containing a vmalloc or vmap address.

    >>> task = find_task(113)
    >>> vmalloc_to_pfn(task.stack)
    (unsigned long)10376

    :param addr: ``void *``
    :return: ``unsigned long``
    """
    return page_to_pfn(vmalloc_to_page(prog, addr))


def _vmap_area_rb_cmp(addr: int, va: Object) -> int:
    if addr < va.va_start.value_():
        return -1
    elif addr >= va.va_end.value_():
        return 1
    else:
        return 0


def _vmap_nodes(prog: Program) -> Object:
    vmap_nodes = prog["vmap_nodes"]
    try:
        return vmap_nodes.read_()
    except ObjectAbsentError:
        # On !SMP and 32-bit kernels, vmap_nodes is initialized to &single and
        # never reassigned. GCC as of version 12.2 doesn't generate a location
        # for vmap_nodes description in that case.
        return prog.variable("single", "mm/vmalloc.c").address_of_()


@takes_program_or_default
def find_vmap_area(prog: Program, addr: IntegerLike) -> Object:
    """
    Return the ``struct vmap_area *`` containing an address.

    >>> find_vmap_area(0xffffa2b680081000)
    *(struct vmap_area *)0xffffa16541046b40 = {
            ...
    }

    :param addr: Address to look up.
    :return: ``struct vmap_area *`` (``NULL`` if not found)
    """
    addr = operator.index(addr)
    # Since Linux kernel commit d093602919ad ("mm: vmalloc: remove global
    # vmap_area_root rb-tree") (in v6.9), vmap areas are split up in multiple
    # red-black trees in separate "nodes". Before that, they're in a single
    # red-black tree.
    try:
        vmap_nodes = _vmap_nodes(prog)
    except KeyError:
        return rb_find(
            "struct vmap_area",
            prog["vmap_area_root"].address_of_(),
            "rb_node",
            addr,
            _vmap_area_rb_cmp,
        )
    else:
        nr_vmap_nodes = prog["nr_vmap_nodes"].value_()
        i = j = (addr // prog["vmap_zone_size"].value_()) % nr_vmap_nodes
        while True:
            vn = vmap_nodes[i]
            va = rb_find(
                "struct vmap_area",
                vn.busy.root.address_of_(),
                "rb_node",
                addr,
                _vmap_area_rb_cmp,
            )
            if va:
                return va
            # As noted in the kernel implementation, the given address may be
            # in a different node than the start address, so we have to loop.
            i = (i + 1) % nr_vmap_nodes
            if i == j:
                return NULL(prog, "struct vmap_area *")


@takes_program_or_default
def for_each_vmap_area(prog: Program) -> Iterator[Object]:
    """
    Iterate over every ``struct vmap_area *`` on the system.

    >>> for va in for_each_vmap_area():
    ...     caller = ""
    ...     if va.vm:
    ...         try:
    ...             sym = prog.symbol(va.vm.caller)
    ...         except LookupError:
    ...             pass
    ...         else:
    ...             caller = f" {sym.name}"
    ...     print(f"{hex(va.va_start)}-{hex(va.va_end)}{caller}")
    ...
    0xffffa2b680000000-0xffffa2b680005000 irq_init_percpu_irqstack
    0xffffa2b680005000-0xffffa2b680007000 acpi_os_map_iomem
    0xffffa2b68000b000-0xffffa2b68000d000 hpet_enable
    0xffffa2b680080000-0xffffa2b680085000 kernel_clone
    ...

    :return: Iterator of ``struct vmap_area *`` objects.
    """
    # Since Linux kernel commit d093602919ad ("mm: vmalloc: remove global
    # vmap_area_root rb-tree") (in v6.9), vmap areas are split up in multiple
    # lists in separate "nodes". Before that, they're in a single list.
    try:
        vmap_nodes = _vmap_nodes(prog)
    except KeyError:
        yield from list_for_each_entry(
            "struct vmap_area", prog["vmap_area_list"].address_of_(), "list"
        )
    else:
        for i in range(prog["nr_vmap_nodes"]):
            yield from list_for_each_entry(
                "struct vmap_area", vmap_nodes[i].busy.head.address_of_(), "list"
            )


def access_process_vm(task: Object, address: IntegerLike, size: IntegerLike) -> bytes:
    """
    Read memory from a task's virtual address space.

    >>> task = find_task(1490152)
    >>> access_process_vm(task, 0x7f8a62b56da0, 12)
    b'hello, world'

    :param task: ``struct task_struct *``
    :param address: Starting address.
    :param size: Number of bytes to read.
    :raises FaultError: if the virtual address
        :ref:`cannot be translated <virtual address translation failures>`
    :raises NotImplementedError: if virtual address translation is :ref:`not
        supported <architecture support matrix>` for this architecture yet
    """
    return _linux_helper_read_vm(task.prog_, task.mm.pgd, address, size)


def access_remote_vm(mm: Object, address: IntegerLike, size: IntegerLike) -> bytes:
    """
    Read memory from a virtual address space. This is similar to
    :func:`access_process_vm()`, but it takes a ``struct mm_struct *`` instead
    of a ``struct task_struct *``.

    >>> task = find_task(1490152)
    >>> access_remote_vm(task.mm, 0x7f8a62b56da0, 12)
    b'hello, world'

    :param mm: ``struct mm_struct *``
    :param address: Starting address.
    :param size: Number of bytes to read.
    :raises FaultError: if the virtual address
        :ref:`cannot be translated <virtual address translation failures>`
    :raises NotImplementedError: if virtual address translation is :ref:`not
        supported <architecture support matrix>` for this architecture yet
    """
    return _linux_helper_read_vm(mm.prog_, mm.pgd, address, size)


def cmdline(task: Object) -> Optional[List[bytes]]:
    """
    Get the list of command line arguments of a task, or ``None`` for kernel tasks.

    >>> cmdline(find_task(1495216))
    [b'vim', b'drgn/helpers/linux/mm.py']

    .. code-block:: console

        $ tr '\\0' ' ' < /proc/1495216/cmdline
        vim drgn/helpers/linux/mm.py

    :param task: ``struct task_struct *``
    :raises FaultError: if the virtual address containing the command line
        :ref:`cannot be translated <virtual address translation failures>`
    :raises NotImplementedError: if virtual address translation is :ref:`not
        supported <architecture support matrix>` for this architecture yet
    """
    mm = task.mm.read_()
    if not mm:
        return None
    arg_start = mm.arg_start.value_()
    arg_end = mm.arg_end.value_()
    return access_remote_vm(mm, arg_start, arg_end - arg_start).split(b"\0")[:-1]


def environ(task: Object) -> Optional[List[bytes]]:
    """
    Get the list of environment variables of a task, or ``None`` for kernel tasks.

    >>> environ(find_task(1497797))
    [b'HOME=/root', b'PATH=/usr/local/sbin:/usr/local/bin:/usr/bin', b'LOGNAME=root']

    .. code-block:: console

        $ tr '\\0' '\\n' < /proc/1497797/environ
        HOME=/root
        PATH=/usr/local/sbin:/usr/local/bin:/usr/bin
        LOGNAME=root

    :param task: ``struct task_struct *``
    :raises FaultError: if the virtual address containing the environment
        :ref:`cannot be translated <virtual address translation failures>`
    :raises NotImplementedError: if virtual address translation is :ref:`not
        supported <architecture support matrix>` for this architecture yet
    """
    mm = task.mm.read_()
    if not mm:
        return None
    env_start = mm.env_start.value_()
    env_end = mm.env_end.value_()
    return access_remote_vm(mm, env_start, env_end - env_start).split(b"\0")[:-1]


def _vma_rb_cmp(addr: int, vma: Object) -> int:
    if addr < vma.vm_start.value_():
        return -1
    elif addr >= vma.vm_end.value_():
        return 1
    else:
        return 0


def vma_find(mm: Object, addr: IntegerLike) -> Object:
    """
    Return the virtual memory area (VMA) containing an address.

    :param mm: ``struct mm_struct *``
    :param addr: Address to look up.
    :return: ``struct vm_area_struct *`` (``NULL`` if not found)
    """
    try:
        # Since Linux kernel commit 524e00b36e8c ("mm: remove rb tree.") (in
        # v6.1), VMAs are stored in a maple tree.
        mt = mm.mm_mt.address_of_()
    except AttributeError:
        # Before that, they are in a red-black tree.
        return rb_find(
            "struct vm_area_struct",
            mm.mm_rb.address_of_(),
            "vm_rb",
            operator.index(addr),
            _vma_rb_cmp,
        )
    else:
        return cast("struct vm_area_struct *", mtree_load(mt, addr))


def for_each_vma(mm: Object) -> Iterator[Object]:
    """
    Iterate over every virtual memory area (VMA) in a virtual address space.

    >>> for vma in for_each_vma(task.mm):
    ...     print(vma)
    ...
    *(struct vm_area_struct *)0xffff97ad82bfc930 = {
        ...
    }
    *(struct vm_area_struct *)0xffff97ad82bfc0a8 = {
        ...
    }
    ...

    :param mm: ``struct mm_struct *``
    :return: Iterator of ``struct vm_area_struct *`` objects.
    """
    try:
        # Since Linux kernel commit 763ecb035029 ("mm: remove the vma linked
        # list") (in v6.1), VMAs are stored in a maple tree.
        mt = mm.mm_mt.address_of_()
    except AttributeError:
        # Before that, they are in a linked list.
        vma = mm.mmap
        while vma:
            yield vma
            vma = vma.vm_next
    else:
        type = mm.prog_.type("struct vm_area_struct *")
        for _, _, entry in mt_for_each(mt):
            yield cast(type, entry)


@takes_program_or_default
def totalram_pages(prog: Program) -> int:
    """Return the total number of RAM pages."""
    try:
        # The variable is present since Linux kernel commit ca79b0c211af63fa32
        # ("mm: convert totalram_pages and totalhigh_pages variables
        # to atomic") (in v5.0).
        return prog["_totalram_pages"].counter.value_()
    except KeyError:
        return prog["totalram_pages"].value_()


@takes_program_or_default
def in_direct_map(prog: Program, addr: IntegerLike) -> bool:
    """
    Return whether an address is within the kernel's direct memory mapping.

    :param addr: address to check
    """
    addr = operator.index(addr)
    start_addr = pfn_to_virt(prog["min_low_pfn"]).value_()
    end_addr = (pfn_to_virt(prog["max_low_pfn"]) + prog["PAGE_SIZE"]).value_()
    return start_addr <= addr < end_addr
